/******************************************************************************
    vcdr.c - VCD file reader.
    These are the utility routines shared by the PLI and stand-alone tools
    Copyright (C) 1998 Christian B. Spear and Synopsys, Inc. 

    $Id: vcdr.c,v 1.17 2008/10/14 23:25:26 spear Exp $                  */
int VCDR_MAJOR_REV = 1;
int VCDR_MINOR_REV = 16;
/*
Edit history
Who When      Version	What
cbs 10/14/08   1.16	Added ERROR macro
cbs  4/17/08   1.15	Support [0:0] subscripts from SystemVerilog
cbs  7/17/02   1.14     Support for creating Verilog shell
cbs  1/03/01   1.14	Add from_vpd
cbs  9/13/00   1.13	Better error messages when running in PLI mode
cbs 12/15/99   1.12	Change time from long to double
cbs 12/10/99   1.12	Better error when MTI abbrevs are unexpectedly found
cbs 10/28/99   1.11	Add support for $vcdread_clocks
cbs  9/28/99   1.10	Add support for MTI style abbreviations
cbs  2/22/99   1.6	Fixed vcdread_comment to get extra tokens
cbs  8/24/98   1.5	Read scalar'ed vectors from XL
			Recognize all $var types, check var width
cbs  7/31/98   1.4	Add $comment
cbs  7/30/98   1.3	Allow $scope and $vars to mix in vcdread_scope
cbs  7/28/98   1.2	Add changes from Xerox for files created by VSS 
cbs  6/29/98   0.4	Lint cleanups
cbs  5/16/98   0.3	Split into PLI & stand-alone versions
cbs  5/09/98   0.2	All pre-PLI stuff in place
cbs  3/22/98   0.1	Major data structures in place
cbs  3/17/98   0.0	Created


Wish list - Ooops, quoted names in a Verilog model, ie. \74LS00 do NOT
have a leading \ in the VCD file.  This changes several assumptions...
In the meantime, don't use quoted names.

$dumpall, $dumpoff, and $dumpon are not supported, and $dumpvars is allowed
only once.

******************************************************************************/


#include "malloc.h"        /* Reduces stress                                 */
#include "string.h"        /* Better string definitions                      */

#define EXT                /* Make VCD data structures local                 */
#include "vcdr.h"          /* VCD Read data structures and globals           */

/* Bit swapping.
   In VSS, the VHDL have subscript [hi:lo], the $var statements
   declare [hi:lo], but the bits in the file are [lo:hi].  When this
   variable is set, any signals with bit subscripts [hi:lo] will have
   their bits swapped when values are read
*/
bool vcdread_vss_bit_swap = {FALSE};

/* This holds the name of all the different $var types                       */
char *var_names[VAR_LAST+1] = {
    "*ERROR*", "reg", "wire", "integer", "parameter", "time", "event", "real",
    "supply0", "supply1", "tri", "triand", "trireg", "tri0", "tri1",
    "wand", "wor"};


extern void exit(int);
extern int isspace(int);

/******************************************************************************
  Procedure - 
      vcdread_comment

  Description - 
      Read $comment ... $end

  Arguments - 
      none

  Returns -
      none
******************************************************************************/
void vcdread_comment()
{
    if (verbose) printf("$comment ");
    fscanf(vcdF, "%s ", prev_word);
    while (strcmp(prev_word, "$end")) {
	if (verbose) printf("%s ", prev_word);
	fscanf(vcdF, "%s ", prev_word);
    }
    if (verbose) printf("$end\n");

    /* Get token AFTER the $comment ... $end */
    fscanf(vcdF, "%s ", prev_word);

} /* vcdread_comment */


/******************************************************************************
  Procedure - 
      expect2

  Description - 
      Read a word of input and compare it to expected value.  Note
      that this only finds a single WORD, not a line.  The word of input
      can be passed in to get consistant error messages.

  Arguments - 
      expected - ptr to string of expected value to be read
      word - [Optional] ptr to string to be compared

  Returns -
      Nothing
******************************************************************************/
void expect2(char *expected, char *word)
{
  /* Get the next word, if needed */
  if (word == 0)
    fscanf(vcdF, "%s ", prev_word);
  else
    strcpy(prev_word, word);

  /* Skip over comments */
  if (!strcmp(prev_word, "$comment")) vcdread_comment();

  /* Check it out! */
  if (strcmp(prev_word,expected)) {
    ERROR("VCDRead", "VCDEXF", "Expected %s, found '%s'", expected, prev_word);
  }

} /* expect2 */



void expect(char *expected)
{
  expect2(expected,0);
}


/******************************************************************************
  Procedure - 
      vcdread_header

  Description - 
      Reads the VCD file header, consisting of the $date, $version, $timescale

  Arguments - 
      None

  Returns -
      Nothing
******************************************************************************/
void vcdread_header()
{
  char line[MAX_LEN], wday[16], month[16], time[16], c;
  int day, year;

  /* VCD files produced by vpd2vcd don't have $date or $version */
  if (!from_vpd) {
  /* Read date header
     $date
     \tFri Apr 17 06:36:50 1998
     $end
     \n 
  */
  expect("$date");
  fscanf(vcdF, "\t%s %s %d %s %d\n", 
      wday, month, &day, time, &year);
  if (verbose)
    printf("File created on %s %s %d %s %d\n",
	   wday, month, day, time, year);
  expect("$end");

  /* Read version.  VSS leaves this blank, so take a peek at the next ch
     $version
     \tChronologic Simulation VCS version 4.0.5
     $end
 */
  expect("$version");
  c = fgetc(vcdF);
  ungetc(c, vcdF);
  if (c != '$') {
      fgets(line, MAX_LEN, vcdF);
      if (verbose) printf("File created by %s",line);
  }
  expect("$end");
  } /* if !from_vpd */


  /* Read timescale
     $timescale
     \t1s
     $end
 */
  expect("$timescale");
  fscanf(vcdF, " %s ", vcd_timescale);
  if (verbose) printf("$timescale = %s\n",vcd_timescale);
  expect("$end");

} /* vcdread_header */

/******************************************************************************
  Procedure - 
      new_scope

  Description - 
      Create and fill in a new scope block

  Arguments - 
      name - ptr to char with scope name (module or instance name)
      Copied into a new malloc'ed piece of memory
      type - "module", "task", "function", "begin", "fork"

  Returns -
      Pointer to new scope block
******************************************************************************/
p_scope new_scope (char *name, char *type)
{
    p_scope scope;
 
    num_scopes = num_scopes + 1;
    scope = (p_scope) malloc(sizeof(s_scope));

    memset((void*)scope, 0, sizeof(s_scope));
    scope->name = name;

    if (!strcmp(type, "module")) {
	scope->type = SCOPE_MODULE;
    }
    else if (!strcmp(type, "function")) {
	scope->type = SCOPE_FUNCTION;
    }
    else if (!strcmp(type, "task")) {
	scope->type = SCOPE_TASK;
    }
    else if (!strcmp(type, "begin")) {
	scope->type = SCOPE_BEGIN;
    }
    else if (!strcmp(type, "fork")) {
	scope->type = SCOPE_FORK;
    }
    else {
    ERROR("VCDRead", "VCDEOF", 
	  "Reading VCD definitions, expected a scope name (module, function, task, begin, fork), instead found '%s'", 
	  type);
    }

    return scope;

} /* new_scope */


/******************************************************************************
  Procedure - 
      abbrev_to_sabbrev

  Description - 
      Calculate the short form of the abbreviation.
      A std abbrev is a string containing just the 94 characters from ! to ~
      A short abbrev hashes the string to:  ! = 1, # = 2, & = 3, ~ = 94,
      !! = 94, #! = 95, &! = 96 ...

	  A MTI abbrev is <9999, so just strip off the < and atoi

  Arguments - 
      abbrev - A string containing the abbrev

  Returns -
      An integer with the hashed abbreviation
******************************************************************************/
int abbrev_to_sabbrev (char *abbrev)
{
    int sabbrev;
	if (use_std_abbrev) {
	  int i;
	  size_t len = strlen(abbrev);

	  if (len > MAX_ABBREV_LEN) {
		ERROR("VCDRead", "VCDABB", "Abbreviation %s is longer than MAX_ABBREV_LEN (%d)",
			    abbrev, MAX_ABBREV_LEN);

		/* MTI VCD files use abbrevs "<999" where 999 is an index */
		/* If this abbrev is of that format, suggest a fix. */
		i = sscanf(abbrev, "<%d", &sabbrev);
		if (i) {
		  printf("The abbreviation seems to be from an MTI VCD file.\n");
#ifdef PLI
		  printf("Try re-running with $vcdread_options(\"mti_abbrev\");\n");
#else
		  printf("Try re-running with the switch -mti_abbrev\n");
#endif
		}
		exit(1);
	  }

	  /* Create the hash'ed abbreviation by adding and multiplying.
		 The abbrev characters ! to ~ are shifted down to 0 .. 93
		 then multiplied by 94
	  */
	  sabbrev = 0;
	  for (i = len -1; i >= 0; i--) {
		if ((abbrev[i] < ABBREV_FIRST_CH) | (abbrev[i] > ABBREV_LAST_CH)) {
		  ERROR("VCDRead", "VCDCHA",
			"Char %0d of abbrev '%s' is out of the acceptable range '%c' (%d.) to '%c' (%d)",
			i, abbrev, ABBREV_FIRST_CH, ABBREV_FIRST_CH, 
			ABBREV_LAST_CH, ABBREV_LAST_CH);
		}
		sabbrev = (sabbrev * SABBREV_RANGE) + (abbrev[i] - ABBREV_FIRST_CH + 1);
	  } /* for i */

	  if (sabbrev > max_sabbrev) 
		max_sabbrev = sabbrev;
	  else if (sabbrev <= 0) {
		ERROR("VCDRead", "VCDCHA", "Sabbrev = %d. out of range in abbrev_to_sabbrev \n",
		      sabbrev);
	  }
	} /* if use_std_abbrev */

	else {
	  sscanf(abbrev, "<%d", &sabbrev);
	  if ((sabbrev < 0) || (sabbrev > MAX_SABBREV)) {
	    ERROR("VCDRead", "VCDAOR",
		  "Sabbrev (%d) out of range 0..MAX_SABBREV (%d) in abbrev_to_sabbrev \n",
		  sabbrev, MAX_SABBREV);
	  }
	  else if (sabbrev > max_sabbrev) 
	    max_sabbrev = sabbrev;
	} /* else use_std_abbrev (ie. MTI abbrev) */
    return sabbrev;

} /* abbrev_to_sabbrev */


/******************************************************************************
  Procedure - 
      vcdread_name

  Description - 
      Read a scope or signal name from the VCD file.
      The name can be unquoted (a-z,0-9,$,_) or quoted, where it starts
      with a \ and ends with a white space, such as "\7400 ", without
      the "   Luckily fscanf("%s") reads the name up to the white space,
      so this routine just needs to add a trailing space.

  Arguments - 
      None

  Returns -
      name - char* to malloc'ed name.
******************************************************************************/
char *vcdread_name ()
{
    char name[MAX_LEN];

    fscanf(vcdF, "%s ", name);
    if (name[0] == '\\') strcat(name, " ");
    return strcpy ((char *)malloc(strlen(name) + 1), name);

} /* vcdread_name */


/******************************************************************************
  Procedure - 
      new_signal

  Description - 
      Create and fill in a new signal block, and link it into 
      the current scope's signal list

  Arguments - 
      name - ptr to char with signal name, vector or scalar
             Copied into a new malloc'ed piece of memory
      scope - ptr to current scope
      abbrev - Signal abbreviation

  Returns -
      Nothing
******************************************************************************/
void new_signal (char *name,
		 p_scope scope,
		 char *abbrev,
		 int width, 
		 int lbit, 
		 int rbit,
		 int var_type)
{
    p_signal signal;

    if (width >= MAX_LEN - 1) {
#ifdef PLI
		  tf_message (ERR_FATAL, "VCDRead", "VCDWSI",
					  "The width (%d.) of signal %s is wider than MAX_LEN - change in vcdr.h and recompile",
	       width, name);
#else
	printf("ERROR (Internal) The width (%d.) of signal %s is wider than %d.\n",
	       width, name, MAX_LEN);
	printf("Please change the constant MAX_LEN in vcdr.h and recompile\n");
	exit(1);
#endif
    }

    signal = (p_signal) malloc(sizeof(s_signal));
    memset((void*)signal, 0, sizeof(s_signal));

    signal->name = name;
    signal->abbrev = strcpy ((char *)malloc(strlen(abbrev) + 1), abbrev);
    signal->sabbrev = abbrev_to_sabbrev(abbrev);

    /* Add this signal to the head of the linked list of signals */
    signal->next_signal = scope->signals;
    scope->signals = signal;

    signal->width = width;
    signal->var_type = var_type;
    signal->parent_scope = scope;

    if (verbose)
	    printf("\t%s ", var_names[var_type]);

    if (width == 1) {
	/* Scalar signal, set bit subscripts                                 */
	signal->lbit = lbit;
	signal->rbit = lbit;
	if (verbose) 
	    printf("[%d] %s;\t// '%s' (%d)\n", 
		   lbit, name, abbrev, signal->sabbrev);
    }
    else {
	/* Vector signal, fill in left and right bit subscripts              */
	signal->lbit = lbit;
	signal->rbit = rbit;
	signal->bit_down = (lbit > rbit);                  /* True for Hi:lo */
	if (verbose)
	    printf("[%d:%d] %s;\t// '%s' (%d)\n",
		   lbit, rbit, name, abbrev, signal->sabbrev);
    }/* if width == 1 */


} /* new_signal */


/******************************************************************************
  Procedure - 
      vcdread_var

  Description - 
      Read a new variable definition from the VCD file.

  Arguments - 
      current_scope - pointer to the current scope
      top_level - True if this is the phantom top level scope

  Returns -
      viod
******************************************************************************/
void vcdread_var(p_scope current_scope)
{
    char type[MAX_LEN], *var_name, tmp_name[MAX_LEN], abbrev[MAX_ABBREV_LEN*4];
    int width, lbit, rbit, var_type, r;
    
    
    /* Now scan through the following lines, looking for signal declarations */
    /* These come in many styles:
       $var wire     32 !    <name> [31:0] $end
       $var tri      32 !    <name> [7] $end
       $var reg       1 "    <name>  $end
      */
    while (!strcmp(prev_word, "$var")) {
      char *ptr;

	fscanf(vcdF, "%s %d %s ", type, &width, abbrev);

	// Convert type to lower case
	ptr = type;
	while (*ptr)
	  *ptr++ = tolower(*ptr);

	var_name = vcdread_name();
	if (width > 1) {
	    /* Vector signal, get left and right bit subscripts
	       Note that integers are 32 bit vectors w/o bit subscripts
	      */
	    if (! fscanf(vcdF, "[%d:%d] ", &lbit, &rbit)) {
		lbit = 31;
		rbit = 0;
	    }
	}
	else {
	  char colon;
	    /* Scalar signal.  Might have a single bit subscript in XL or double in SV */
	    lbit = 0;
	    rbit = 0;
	    r = fscanf(vcdF, "[%d", &lbit);

	    /* Look for ":r]" */
	    colon = getc(vcdF);
	    if (colon == ':')
	      fscanf(vcdF, "%d", &rbit);
	    else
	      ungetc(colon, vcdF); /* Oops, not a colon */
	    fscanf(vcdF, "] ");

	    /* In bit blasting mode, change 'vec [0]' to 'vec_0' */
	    if (bit_blast && r>0) {
	      sprintf(var_name, "%s_%d", var_name, lbit);
	      //	      strcpy(var_name, tmp_name);
	    }
	} /* if (width > 1) */
	
	/* Find the matching $var type */
	for (var_type = VAR_LAST; var_type > VAR_ERROR; var_type--)
	    if (strcmp(type, var_names[var_type]) == 0) break;
	if (var_type == VAR_ERROR) {
	    printf("WARNING, unregonized $var type '%s', using 'wire'\n", type);
	    var_type = VAR_WIRE;
	}

	new_signal (var_name, current_scope, abbrev, 
		    width, lbit, rbit, var_type);
	
	expect("$end");
	fscanf(vcdF, "%s ", prev_word);

    } /* while $var */

} /* vcdread_var */



/******************************************************************************
  Procedure - 
      vcdread_scope

  Description - 
      Read a new scope from the VCD file, building a list of signals
      and sub-scopes / instances

  Arguments - 
      parent_scope - pointer to the parent of this new scope
      top_level - True if this is the phantom top level scope
      The head of the list of scopes is called Phantom, and its instances are
      all the true top level scopes in the model

  Returns -
      Address of scope block
******************************************************************************/
p_scope vcdread_scope(p_scope parent_scope, bool top_level)
{
    char *scope_name, scope_type[MAX_LEN];
    p_scope current_scope;
    bool done;

    if (top_level) {
	/* This is the top level "phantom" scope that has as its
	   children all of the real top level scopes.  Note that all
	   other names used in this program point to malloc'ed name,
	   while this one is a constant.  */
	current_scope = new_scope("Phantom", "module");
    }
    else {
	/* Read scope.  The $scope was already read by the calling rtn
	   Scope type is module, function, task, begin, or fork
	   $scope <type> <name> $end
	  */
	fscanf(vcdF, "%s ", scope_type);
	scope_name = vcdread_name();
	if (verbose)
	    printf("    %s %s;\n", scope_type, scope_name);
	expect("$end");
	fscanf(vcdF, "%s ", prev_word);
	current_scope = new_scope(scope_name, scope_type);
    }
    current_scope->parent_scope = parent_scope;
    
    done = FALSE;
    while (!done) {
      if (! strcmp(prev_word, "$var")) {
	      /* Look for variable definitions */
	      vcdread_var(current_scope);
	  }

      else if (! strcmp(prev_word, "$scope")) {
	      /* Read the instance and link it to list of instances */
	      p_scope instance = vcdread_scope(current_scope, 0);
	      instance->next_sibling = current_scope->first_instance;
	      current_scope->first_instance = instance;
	  }

      else
	      done = TRUE;
  } /* while !done */


    /* Note that at this point prev_word contains the latest read from the
       file, but it is not $var or $scope.  Expect $enddefinition because
       we are at the end of defining the top level scope, or just $upscope
       because we are at the end of this current scope
       */
    if (top_level) {
        /* Top level scope ends with 
	   $enddefinitions $end 
           */
        expect2("$enddefinitions", prev_word);
        expect("$end");
	return current_scope;
    }
    else {
	expect2("$upscope", prev_word);
	expect("$end");
	fscanf(vcdF, "%s ", prev_word);
	if (verbose)
	    printf("    end%s // %s\n", scope_type, scope_name);
    }

    return current_scope;

} /* vcdread_scope */


/******************************************************************************
  Procedure - 
      traverse_scopes

  Description - 
      Recursively traverse the tree of scopes, building the abbrev table
      Make a list of all signals with the same abbreviation

  Arguments - 
      scope - Current scope to traverse
      level - Level in hierarchy, top = 0

  Returns -
      Number of scopes found - for error check checking
******************************************************************************/
int traverse_scopes(p_scope scope, int level)
{
    p_scope current_scope = scope;
    p_signal signal = current_scope->signals;
    int scopes_found;

    scopes_found = 1;

    if (verbose)
	printf ("Traversing scope %s, level = %d\n", current_scope->name, level);

    /* Traverse all the signals in this scope */
    while (signal) {
	if (verbose && 0)
	    printf("\t Found signal %s '%s', %d. (0x%x)\n", 
		   signal->name, signal->abbrev, signal->sabbrev, (int) signal);

	if ((signal->sabbrev < 0) | 
	    (signal->sabbrev > max_sabbrev)) {
	    printf("ERROR (Internal) short abbrev %d out of range 0..%d\n",
		   signal->sabbrev, max_sabbrev);
	    printf("For signal %s.%s\n", 
		   current_scope->name, signal->name);
	    exit(1);
	}
	/* Link this signal to the list of all signals with the same abbrev  */
	signal->next_abbrev = sabbrev_table[signal->sabbrev];
	sabbrev_table[signal->sabbrev] = signal;

	/* Now move on to the next signal in this scope                      */
	signal = signal->next_signal;
    } /* while (signal) */
	

    /* Traverse all the instances under this scope */
    if (scope->first_instance) {
	scopes_found = scopes_found + 
	  traverse_scopes(scope->first_instance, level + 1);
    }
    /* Lastly traverse the siblings of this scope, ie. all scopes with
       the same parent as this scope */
    current_scope = scope->next_sibling;
    if (current_scope) {
	scopes_found = scopes_found + 
	  traverse_scopes(current_scope, level);
	current_scope = current_scope->next_sibling;
    }

    return scopes_found;

} /* traverse_scopes */


/******************************************************************************
  Procedure - 
      generate_verilog_modules

  Description - 
      Recursively traverse the tree of scopes, creating a Verilog shell

  Arguments - 
      scope - Current scope to traverse
      hname - hierarchical name so far...
      level - depth in hierarchy      

  Returns -
      Nothing
******************************************************************************/
void generate_verilog_modules(p_scope scope, char *hname, int level)
{
    p_scope instance, current_scope = scope;
    p_signal signal = current_scope->signals;
    int instances = 0;
    char iname[MAX_LEN];

    switch (level) {
    case 0:
      /* This is the Phantom scope */
      iname[0] = '\0';
      break;
    
    case 1:
      /* This is a top level module */
      strncat(iname, current_scope->name, MAX_LEN);
      fprintf(verilogF, "module %s;\n", iname);
      break;

    default:
      /* Just some instance in the design */
      strncpy(iname, hname, MAX_LEN);
      strncat(iname, "_", MAX_LEN);
      strncat(iname, current_scope->name, MAX_LEN);
      fprintf(verilogF, "module %s;\n", iname);
    } /* switch (level) */


    /* Traverse all the signals in this scope */
    while (signal) {
      if (signal->width == 1)
	fprintf(verilogF, "%s %s;\n",
		var_names[signal->var_type], signal->name);
      else
	fprintf(verilogF, "%s [%d:%d] %s;\n",
	      var_names[signal->var_type], signal->lbit, signal->rbit, 
		signal->name);

	/* Now move on to the next signal in this scope                      */
	signal = signal->next_signal;
    } /* while (signal) */
	

    /* Print all the instances in this module */
    if (level > 0) {
      instance = scope->first_instance;
      while (instance) {
	fprintf(verilogF, "%s_%s instance_%d;\n", 
		iname, instance->name, instances++);
	instance = instance->next_sibling;
      }

      fprintf(verilogF, "endmodule\n\n");
    }

    /* Traverse all the instances under this scope */
    if (scope->first_instance) {
      generate_verilog_modules(scope->first_instance, iname, level+1);
    }

    /* Lastly traverse the siblings of this scope, ie. all scopes with
       the same parent as this scope */
    current_scope = scope->next_sibling;
    if (current_scope) {
      generate_verilog_modules(current_scope, hname, level);
      current_scope = current_scope->next_sibling;
    }

} /* generate_verilog_modules */


/******************************************************************************
  Procedure - 
      check_sabbrev_table

  Description - 
      Check the abbrev table for internal errors

  Arguments - 
      none

  Returns -
      nothing
******************************************************************************/
void check_sabbrev_table()
{
    int n;
    if (sabbrev_table[0] && use_std_abbrev) {
	printf("ERROR (internal), sabbrev_table[0] = %d, should be 0\n",
	       sabbrev_table[0]);
	exit(1);
    }

    for (n = 1; n <= max_sabbrev; n++) {
        p_signal signal = sabbrev_table[n];
	if (verbose && 0) {
	    printf ("sabbrev[%d] = 0x%x\n", n, (int) signal);
	    if (signal == 0) 
		printf("Info (Internal) Empty abbrev table entry %d\n", n);
	}
	while (signal) {
 	    if (signal == signal->next_abbrev) {
	        printf("ERROR (Internal) Next_abbrev of %s / %s is itself\n",
		       signal->name, signal->abbrev);
		exit(1);
	    }
 	    if (signal->sabbrev != n) {
	        printf("ERROR (Internal) Short_abbrev (%d.) does not match sabbrev table number (%d.)\n",
		       signal->sabbrev, n);
	        printf("for signal '%s'\n", signal->name);
		exit(1);
	    }
	    signal = signal->next_abbrev;
	} /* while */
    }/* for n */

} /* check_sabbrev_table */


/******************************************************************************
  Procedure - 
      make_sabbrev_table

  Description - 
      All the variable definitions & abbreviations have been read.
      Now create and fill in the abbrev lookup table
      This table can be indexed by hashing the abbrev into a 
      short_abbrev.  The each entry in the table is the head of a list of 
      signals which have the same abbreviation.

  Arguments - 
      none

  Returns -
      nothing
******************************************************************************/
void make_sabbrev_table()
{
    int n;

    if (verbose) 
        printf("There are %d short abbrevs\n", max_sabbrev);

    if (max_sabbrev > MAX_SABBREV) {
	printf("ERROR (Internal) max_sabbrev (%d.) > MAX_SABBREV (%d.)\n",
		   max_sabbrev, MAX_SABBREV);
	exit(1);
    }

    sabbrev_table = (p_signal*) malloc(sizeof(p_signal) * (max_sabbrev + 1));
    if (!sabbrev_table) {
	printf ("ERROR (Internal) in malloc abbrev table for %d. entries\n",
		max_sabbrev);
	exit(1);
    }
    memset(sabbrev_table, 0, sizeof(p_signal) * (max_sabbrev + 1));

    /* Traverse the scopes, building the table, and count scopes */
    n = traverse_scopes(scope_head, 0);
    if (n != num_scopes) {
        printf("ERROR (Internal) Number of scopes found by new_scopes %d, ",
	       n);
        printf("does not match those in traverse_scopes %d\n", num_scopes);
	exit(1);
    }

    // Generate dummy Verilog code?
    if (gen_verilog) {
      verilogF = fopen("shell.v","w");
      if (!verilogF) {
	printf("ERROR, unable to open the Verilog shell file shell.v\n");
	exit (1);
      }

      fprintf(verilogF, "`timescale %s/%s\n", vcd_timescale, vcd_timescale);
      generate_verilog_modules(scope_head, NULL, 0);
      fclose(verilogF);
    }

    check_sabbrev_table();

} /* make_sabbrev_table */

/******************************************************************************
  Procedure - 
      vcdread_definitions

  Description - 
      Reads module scopes.  I expect to find at least one!

  Arguments - 
      none

  Returns -
      nothing
******************************************************************************/
void vcdread_definitions()
{
    /* Read module scopes
       $scope module name $end
       */
    expect("$scope");
    scope_head = vcdread_scope((p_scope) 0, 1);

    /* Process the signal abbrevs and create lookup table */
    make_sabbrev_table();

} /* vcdread_definitions */


/******************************************************************************
  Procedure - 
      prepend_scope_name

  Description - 
      Prepends the full hiearchical path for a signal into the 
      beginning of a string.
      Don't print the top-most scope name, it is "Phantom" - shhh!

  Arguments - 
      scope - The scope to be printed, except as above
      str - Pointer to a string being built up

  Returns -
      Nothing
******************************************************************************/
void prepend_scope_name(p_scope scope, char *str)
{
   if (scope->parent_scope == scope_head) {
       /* This is the top "real" scope, its parent is the Phantom */
       strcpy (str, scope->name);
   }
   else {
       /* Not at the top yet.  Get the upper scopes, then append this name */
       prepend_scope_name(scope->parent_scope, str);
       strcat (str, ".");
       strcat (str, scope->name);
   }

} /* prepend_scope_name */


/******************************************************************************
  Procedure - 
      get_full_name

  Description - 
      Pieces together the full hiearchical name of a signal

  Arguments - 
      signal - The signal to be printed

  Returns -
      Pointer to buffer holding full name. 
******************************************************************************/
char *get_full_name(p_signal signal)
{
   static char full_name[MAX_LEN*20];

   prepend_scope_name(signal->parent_scope, full_name);
   strcat(full_name, ".");
   strcat(full_name, signal->name);

   return full_name;

} /* get_full_name */


/******************************************************************************
  Procedure - 
      abbrev_to_signal

  Description - 
      Finds the first signal matching an abbreviation.

  Arguments - 
      abbrev - Abbreviation for signal

  Returns -
      signal - The signal at the head of the list

******************************************************************************/
p_signal abbrev_to_signal (char *abbrev)
{
    p_signal signal = sabbrev_table[abbrev_to_sabbrev(abbrev)];

    if (!signal) {
        ERROR("VCDRead, VCDNoAb", "No abbreviation found for '%s'\n", abbrev);
    }

    return signal;

} /* abbrev_to_signal */


/******************************************************************************
  Procedure - 
      split_hierarchical_name

  Description - 
      Breaks a hierarchical name into pieces - instance & signal names.
      Here are some enlightening examples for the reader:

      str      split    rest
      a        a        0
      i.b      i        a
      \1       \1       0
      \1 .a    \1       a

  Arguments - 
      str - string containing the full hierarchical name
      split - address of string to write the instance or signal name

  Returns -
      char *rest - ptr to the rest of string, after the name and separator .
      Returns 0 if there was no more name, ie. this was a signal name

******************************************************************************/
char *split_hierarchical_name (char *str, char *split)
{
    int i, len = strlen (str);

    /* First, get the leading part of the name*/

    /* Look for an escaped name */
    if (str[0] == '\\') {
        /* Escaped names start with \ and end with white space */
        i = 2;
	while (!isspace(str[i])) i++;
	strncpy(split, str, ++i);  /* Move past trailing space */
	split[i] = '\0';
    }
    else {
        /* This is not an escaped name */
        i = 1;
	while (i < len && str[i] != '.') i++;
	strncpy (split, str, i);
	/* Wish list - does strncpy copy a null? */
	split[i] = '\0';
    }


    /* Return a pointer to the rest of the string, after the . */
    /* If we are now at the end of the string, then return NULL */
    if (i == len)
        return NULL;      /* Split contained signal name, we are at end */
    else
        return &str[i+1]; /* Split has instance name, return rest */

} /* split_hierarchical_name */


/******************************************************************************
  Procedure - 
      str_to_signal

  Description - 
      Given a full hierarchical name, find the matching signal in the
      VCD file.  This is done by traversing the scope tree using the
      instance names, then matching the signal name.

      This could be done faster if the scopes and signals were sorted,
      but let's get real - typical usage will be no more than 1000
      signals being looked up.

      Hey - I've a great idea.  If the routine's performance needs
      improvement, how about this: Most signals looked up will be in
      the same instance.  Why not move the found signals to the end of
      the signal list under scope block, so they would not have to be
      compared again?  This would require adding a tail pointer to a
      scope block.

  Arguments - 
      hname - string containing the full hierarchical name
      split - address of string to write the instance or signal name

  Returns -
      char *rest - ptr to the rest of string, after the name and separator .
      Returns 0 if there was no more name, ie. this was a signal name

******************************************************************************/
p_signal str_to_signal (char *hname)
{
    char *rest, split[MAX_LEN];
    p_signal signal;
    p_scope scope;

    if (verbose)
      printf("str_to_signal - looking up '%s'\n", hname);

    /* Get the top-most module name from the hierarchical name */
    rest = split_hierarchical_name (hname, split);
    if (rest == NULL) {
        ERROR("VCDRead", "VCDFull",
	      "VCD signal names must have full hierarchy such as 'top.dut.in1'\nYou gave just '%s'\n", hname);
    }


    scope = scope_head;    /* Start at the top of the VCD scopes */
    while (rest) {

        /* Look for the scope name under the current scope */
        scope = scope->first_instance;
	while (scope && strcmp (scope->name, split) ) 
	    scope = scope->next_sibling;

	if (!scope) {
	  ERROR("VCDRead", "VCDInF",
		     "Instance '%s' not found for VCD signal '%s'\n",
		     split, hname);
	}
	/* Get the next hierarchical name */
	rest = split_hierarchical_name (rest, split);
    }


    /* Okay, now we have the scope block, match the signal within */
    if (! (signal = scope->signals) ) {
      ERROR("VCDRead", "VCDNoSG",
	    "No signals in scope '%s' for VCD signal '%s'\n",
	    scope->name, hname);
    }

    while (signal && strcmp (signal->name, split))
        signal = signal->next_signal;

    if (!signal) {
      ERROR("VCDRead", "VCDSigNF", "VCD signal '%s' was not found in VCD file.\n", hname);
    }

    return signal;

} /* str_to_signal */


/******************************************************************************
  Procedure - 
      get_vcd_time

  Description - 
      Reads the next simulation time from the VCD file.
      Wish list - handle time correctly as a 64 bit unsigned integer.
      For now it is treated as a 64-bit double

  Arguments - 
      time - Address of double to store time

  Returns -
      Long with Next stimulus time
******************************************************************************/
void get_vcd_time(double *time)
{
  char ch;

  if (feof(vcdF)) {
    ERROR("VCDRead", "VCDEOF", "unexpected EOF when reading time from VCD file.");
  }

  if (fscanf(vcdF, "%lf ", time) != 1) {
    ERROR("VCDRead", "VCDEOF", "Problem reading time from VCD file");
  }

} /* get_vcd_time */
