/******************************************************************************
    vcdr.h - VCD file reader declaration
    Copyright (C) 1998 Christian B. Spear and Synopsys, Inc.


$Id: vcdr.h,v 1.13 2008/10/14 23:27:13 spear Exp $
*/
#define VCDR_H_MAJOR_REV 1
#define VCDR_H_MINOR_REV 12
/*
Edit history
Who When      Version	What
cbs 10/14/08  1.12	Add start_after_0 and ERROR macro
cbs  7/17/02  1.11	Add gen_verilog - version numbers must have glitched
cbs  1/03/01  1.14	Add from_vpd
cbs  8/08/00  1.13      Add support for real numbers
cbs 12/17/99  1.12	Add bit blasting
cbs 10/28/99  1.11	Add $vcdread_clocks
cbs  9/28/99  1.8	Add support for MTI style abbreviations
cbs  9/15/98  1.6	Add support for pattern files
cbs  8/26/98  1.5	Added $var types, increase MAX_LEN
cbs  7/30/98  1.4	Added TRUE, FALSE definitions
cbs  5/16/98  0.0	Created

******************************************************************************/

#ifdef PLI
#include "vcsuser.h"       /* Base Verilog PLI definitions */
#include "acc_user.h"      /* Access routine definitions */
#define ZERO_DELAY 0
#define UNIT_DELAY 1
#define INERTIAL_DELAY 0
#else
/* Useful stuff defined in VCS files */
#define bool int           /* Defined in acc_user.h */
#define true    1
#define false   0
#define TRUE    1
#define FALSE   0

#endif

#include <stdio.h>


/*****************************************************************************/
/* Constant declarations */
/* Abbreviation information
   A single abbreviation can be used for several signals.  For example,
   if a signal is connected to a port, both will have the same abbreviation.
   foo i1 (.in_port(driver));
   Abbreviations consist of the 94 printable characters from ! to ~
   I've put in a 3 character limit for now.... Gives us 830k abbreviations
   Note that sometimes there can be holes in the abbreviation list,
   sometimes ones are skipped.  #5, #10, ...

   Having said all that, it would appear that MTI uses abbrevs of the form 
   '<999999', ie. '<' followed by a unique decimal number. 
   */

#define MAX_LEN 4010
#define ABBREV_FIRST_CH '!'
#define ABBREV_LAST_CH '~'
#define SABBREV_RANGE (ABBREV_LAST_CH - ABBREV_FIRST_CH + 1)
#define MAX_ABBREV_LEN 3
#define MAX_SABBREV SABBREV_RANGE * SABBREV_RANGE * SABBREV_RANGE

#ifdef PLI
#define printf io_printf
#endif

/*****************************************************************************/
/* Signal information.  This is for both scalars and vectors */
#define VAR_ERROR     0       /* Uninitialized! */
#define VAR_REG       1       /* Put the most common types first */
#define VAR_WIRE      2
#define VAR_INTEGER   3       /* VCS doesn't use this type, makes 32 bit reg */
#define VAR_PARAMETER 4       /* VCS never writes these values */
#define VAR_TIME      5
#define VAR_EVENT     6
#define VAR_REAL      7
#define VAR_SUPPLY0   8
#define VAR_SUPPLY1   9
#define VAR_TRI       10
#define VAR_TRIAND    11
#define VAR_TRIREG    12
#define VAR_TRI0      13
#define VAR_TRI1      14
#define VAR_WAND      16
#define VAR_WOR       16
#define VAR_LAST      16

typedef struct t_signal
{
  char *name;              /* Full name of signal                            */
  struct t_signal *next_signal; /* Next signal in this scope                 */
  struct t_scope *parent_scope; /* Parent scope of this signal               */
  char *abbrev;            /* Abbreviation name for this signal              */
  int sabbrev;             /* Hashed version of the abbreviation             */
  struct t_signal *next_abbrev; /* Next signal with this abbreviation        */
  int width, lbit, rbit;   /* Width of signal, [lbit:rbit]                   */
  bool bit_down;           /* Direction of bits, [hi:lo]=True, [lo:hi]=False */
#ifdef PLI
  int argn;                /* Argument number for use in tf_strdelputp       */
  char *instance;          /* Instance number where signal was given         */
  int delay;               /* Delay when applying signal, 0 for clocks       */
#else
  bool print;              /* Print this signal because of -s switch         */
  union {                  /* Last value read, for pattern files.            */
  char* vector;            /* Pointer to last value read for vector signals  */
  char scalar;             /* Character holding last value for scalar sigs   */
  double real;             /* Double holding real value                      */
} last_val;
  struct t_signal *next_psignal; /* Next signal in pattern list              */
#endif
  int var_type;            /* Type of signal: wire, register, etc.           */
} s_signal, *p_signal;


/*****************************************************************************/
/* Scope information
   A scope can be a module, task, function, or named block (begin or fork)
   At the top level scope, the VCD file gives the module name, but lower
   down, the VCD just has the instance name.
   The hierarchical tree is turned on its side.  Instead of a parent
   pointing to multiple instances (children), the parent points to the first
   instance, or left-most.  The other instances (siblings of the first) are
   pointed to by the next pointer.
*/
enum SCOPE_TYPE {
    SCOPE_ERROR,
    SCOPE_MODULE,
    SCOPE_FUNCTION,
    SCOPE_TASK,
    SCOPE_BEGIN,
    SCOPE_FORK
};


typedef struct t_scope
{
  char *name;              /* Name of module or instance */
  struct t_scope *next_sibling;   /* Next sibling of this scope */
  struct t_signal *signals;       /* List of signals in this scope */
  struct t_scope *first_instance; /* First instance in this scope */
  struct t_scope *parent_scope;   /* Parent of this scope */
  enum SCOPE_TYPE type;    /* module, task, function, begin, or fork */
} s_scope, *p_scope;



/*****************************************************************************/
/* Global declarations                                                       */

EXT char VCD_name[MAX_LEN];/* Name of VCD file being read                    */
EXT FILE *vcdF;            /* VCD file being read                            */
EXT char *vcdr_eof_inst;   /* Pointer to instance of $vcdread_eof            */
EXT int verbose;           /* Should we be verbose? You decide!              */
EXT 
char prev_word[MAX_LEN];   /* Previous word read from file                   */
                           /* This is a hack in read_scope to get "$upscope" */

EXT int max_sabbrev;       /* Largest short abbrev found                     */

EXT int num_scopes;        /* Number of scopes created by new_scope          */
EXT p_signal signal_head;  /* Head of list of signals                        */
EXT p_scope scope_head;    /* Head of list of scopes                         */

EXT 
p_signal *sabbrev_table;   /* Table of head of list of signals with common   */
                           /* abbreviations - for fast lookups               */

EXT char vcd_timescale[64];/* Timescale read from the file, used by PLI      */

/*****************************************************************************/
/* Pattern file declarations */
EXT char pat_file[MAX_LEN];/* Name of pattern file being written             */
EXT FILE *patF;            /* Pattern file being written                     */
EXT bool pattern;          /* True if we are writing pattern files           */
EXT char name_file[MAX_LEN]; /* Name of file with list of signals            */
EXT FILE *nameF;           /* Signal name file being read                    */

enum STROBE_ENUM {STROBE_NONE,    /* Uninitialized strobe type               */
		  STROBE_TIME,    /* Strobe on time change                   */
		  STROBE_SIGNAL}; /* Strobe on signal change                 */
EXT enum STROBE_ENUM strobe_type; /* What should we strobe on?
      */
EXT bool strobe_time;      /* T: if strobe on any time change, F: on signal  */
EXT char strobe_name[MAX_LEN]; /* Name of signal used to strobe patterns     */
EXT p_signal strobe_signal, /* Signal structure used for strobe              */
    head_pnames, tail_pnames; /* Linked list of signals in name file         */

/* It appears that MTI uses abbrevs of the form '<999999', ie. '<'
   followed by a unique decimal number.  The following switches between
   the MTI and the standard abbreviations. */
EXT bool use_std_abbrev;

/* When a vector is broken into bits in the VCD file, ie. vec[0], vec[1],
   bit blasting converts the name into a more usable format: vec_0, vec_1    */
EXT bool bit_blast;

/* VCD files produced by vpd2vcd are missing some of the header info.        */
EXT bool from_vpd;

/* Verilog shell file with just variable declarations */
EXT bool gen_verilog;
EXT FILE *verilogF;

/* Allow VCD file that starts after time 0.  Beware - this has not been tested */
EXT bool start_after_0;

#ifdef PLI
#define ERROR(abb1,abb2,...) \
 do {tf_message (ERR_ERROR, abb1, abb2, __VA_ARGS__); \
     tf_dofinish(); } while (0)
#else
#define ERROR(abb1,abb2,...) \
 do {printf("ERROR - "); \
     printf(__VA_ARGS__); \
     exit(1); } while (0)
#endif
