/******************************************************************************
    vcdr_pli.c - VCD file reader 
    These are the routines for the PLI version of the reader
    Copyright (C) 1998, 1999, 2000, 2001, 2008 Christian B. Spear and Synopsys, Inc.

    $Id: vcdr_pli.c,v 1.18 2008/10/27 12:14:13 spear Exp $
*/
#define VCDR_PLI_MAJOR_REV 1
#define VCDR_PLI_MINOR_REV 18
/*
Edit history
Who When      Version	What
cbs 10/14/08   1.18	@Add $vcdread_option("start_after_0"), ERROR macro
cbs  5/04/01   1.17	Allow for different timescales using scaling
cbs  2/27/01   1.16	Allow for real registers as destinations
cbs  9/13/00   1.15	Check for $vcdread_options after $vcdread
cbs  8/08/00   1.14     Add support for real numbers
cbs  5/04/00   1.13b    Change to logical compares in *_pairs_check
cbs 12/15/99   1.12	Add support for MTI VCD signal values pD,pL,etc.
			Change time from long to double
cbs 10/28/99   1.11	Add $vcdread_clocks
cbs  9/28/99   1.10	Add support for MTI style abbreviations
cbs  9/15/99   1.9	Better warning message for multiple $vcdread_eof calls
cbs  5/11/99   1.8	Better check for $ commands at end of file
kg   6/16/99		Fix bit reversing with VSS
cbs  4/15/99   1.7	Better error message on vcdread_signal_pairs,
			Check time unit / precision
			Add $vcdread_eof
cbs  9/14/98   1.6	Really pad the bits when reversing for VSS
cbs  8/04/98   1.5	Pad the bits when reversing for VSS
cbs  7/31/98   1.4	Check bit ordering
cbs  7/30/98   1.3	Replace acc calls with tf
cbs  7/28/98   1.2	Add changes from Xerox for files created by VSS 
cbs  5/16/98   0.3	Split into PLI & stand-alone versions
cbs  5/09/98   0.2	All pre-PLI stuff in place
cbs  3/22/98   0.1	Major data structures in place
cbs  3/17/98   0.0	Created

Wish list 
   - Change all error messages from printf to tf_message / tf_error

   - What happens if $vcdread starts after time 0?  Offset VCD times?
   Give me a call if you have a preference.  Until then, choke on it.

   - Can only have one $vcdread task.  Use tf_get_instance() to find the
   instance pointer and save the arg number.

******************************************************************************/



#include "malloc.h"        /* Reduces stress */
#include "string.h"        /* Better string definitions */

#define EXT extern
#include "vcdr.h"          /* VCD Read data structures and globals */
#include "math.h"          /* pow, no kung */

extern p_signal abbrev_to_signal (char *abbrev);
extern p_signal str_to_signal (char *hname);
extern vcdread_header();
extern vcdread_definitions();
extern vcdread_comment();
extern void get_vcd_time(double *);

extern bool vcdread_vss_bit_swap;

/* Arguments passed into a Verilog system task or function */
#define ARG_RET 0
#define ARG1 1
#define ARG2 2
#define ARG3 3
#define ARG4 4


/* Global declarations */

double vcd_time;
double vcd2mod_scale;          /* Scale VCD delays to module delays    */
int signal_delay = ZERO_DELAY; /* Default to using 0 delay for signals */

static bool vcdread_called = FALSE;


/******************************************************************************
 $vcdread routines
******************************************************************************/

/******************************************************************************
  Procedure - 
      vcdr_check

  Description - 
      Checks arguments for $vcdread calls

  Arguments - 
      none used

  Returns -
      void
******************************************************************************/
void vcdr_check()
{
    if (tf_nump() != 1) {
      tf_message (ERR_ERROR, "VCDRead", "VCDFL1",
		  "$vcdread has only one argument - the file name");
      tf_dofinish();
    }

    /* Wish list - allow file name to be stored in a reg */
    if (tf_typep(ARG1) != tf_string) {
      tf_message (ERR_ERROR, "VCDRead", "VCDSTR",
		  "Bad argument to $vcdread, must be string");
      tf_dofinish();
    }

    /* Initialize this to 0 so $vcdread_signals can tell if $vcdread
       has been called */
    vcdF = (FILE *) 0;

    /* Initialize this to use standard abbrevs, not MTI style */
    use_std_abbrev = TRUE;

    /* Initialize this to turn off bit blasting */
    bit_blast = FALSE;

} /* vcdr_check */


/******************************************************************************
  Procedure - 
      vcdr_call

  Description - 
      Opens VCD file and schedules stimulus to be applied at the end of the
      current time slot.

  Arguments - 
      none used

  Returns -
      void
******************************************************************************/
void vcdr_call()
{
   char ch;

  /* Make sure $vcdread is only called once during simulation */
  if (vcdread_called) {
    tf_message (ERR_ERROR, "VCDRead", "VCDTWC", 
		"$vcdread was called more than once during simulation");
    return;
  }
  else vcdread_called = TRUE;

  /* What happens if $vcdread starts after time 0?  Offset VCD times?
     Give me a call if you have a preference.  Until then, choke on it.
     Wish list - use tf_getlongtime instead.
  */
  if (tf_gettime()) {
    tf_message (ERR_ERROR, "VCDRead", "VCDSCV",
		"$vcdread was called after time 0, call Chris Spear!\n");
    tf_dofinish();
  }

  /* Okay so far, now get VCD file name. */
  strcpy (VCD_name, tf_getcstringp(ARG1));
  vcdF = fopen(VCD_name, "r");
  if (!vcdF) {
    tf_message (ERR_ERROR, "VCDRead", "VCDOPE",
		"Problem opening VCD file '%s'\n", VCD_name);
    tf_dofinish();
  }


  /* Okay, now we have the file open.  Read the header, signal defs,
     and the the first time from the VCD file - should be 0.  
     Wish list - decide how to handle initial times > 0 */
  vcdread_header();
  vcdread_definitions();

   /* #0 is optional in XL dump files */
   ch = fgetc(vcdF);
   if (ch == '#')
       get_vcd_time(&vcd_time);
   else {
       /* If no #, then maybe a $dumpvars? */
       vcd_time = 0.0;
       ungetc(ch, vcdF);
   }

   /* Allow VCD files that start after time 0 - not well tested */
  if ((vcd_time != 0.0) && !start_after_0) {
    tf_message (ERR_ERROR, "VCDRead", "VCDTM0",
	"Initial time %0.0f in VCD file '%s' should be 0.\nOr, call $vcdread_options(\"start_after_0\") before $vcdread(\"%s\")\n",
	   vcd_time, VCD_name, VCD_name);
    tf_dofinish();
  }

  expect("$dumpvars");

  /* Have vcdr_misc called at the end of the time slot.  The user
     should have the $vcdread_signals call done immediately after the
     $vcdread call */
  tf_synchronize();

} /* vcdr_call */


/******************************************************************************
  Procedure - 
      vcdread_changes

  Description - 
      Read signal changes from the VCD file until a delay is found, or
      an error.  Value changes are read from the file and applied to
      simulation.

      Wish list - when we know there are no more stimulus points being
      defined, ie. no more calls to $vcdread_signals, optimize the VCD
      signal lists by deleting the signals which are not drivers.
      This may give a speedup - but then again, it might not!

  Arguments - 
      none

  Returns -
      !feof(vcdF) - True if more data in the file, false if end of file
******************************************************************************/
bool vcdread_changes()
{
  char value[MAX_LEN],        /* Vector value read from the file */
    abbrev[MAX_ABBREV_LEN*4], /* Signal abbreviation */
    ch, level;
  p_signal signal;
  int i, j;
  static bool dollar_end_found = FALSE; /* True if $end found - error check */


  ch = fgetc(vcdF);
  while(!feof(vcdF)) {
    switch (ch) {

    case '0': 
    case '1': 
    case 'z': 
    case 'x': 
    case 'Z': 
    case 'X': 
      /* Found a scalar value - [01xz]abbrev */
      fscanf(vcdF, "%s ", abbrev);
      signal = abbrev_to_signal(abbrev);

      /* Schedule transistions on all signals with this abbreviation,
	 and which drive registers in the simulation.  */
      while (signal) {
	if (signal->argn) {
	  if (verbose){
	    io_printf("VCD Driving %s (%s) = %c;\n", 
		   get_full_name(signal), abbrev, ch);
	  }
	  value[0] = ch;
	  value[1] = '\0';
	  tf_istrdelputp(signal->argn, 1, 'b', value, 
			 signal->delay, INERTIAL_DELAY, signal->instance);
	}
	signal = signal->next_abbrev;
      }
      break; /* 01zxXZ */


    case '$':
      /* Is this a $comment or $end ? */
      ungetc(ch, vcdF);
      fscanf(vcdF, "%s ", prev_word);
      if (!strcmp(prev_word, "$comment")) vcdread_comment();
      else {
	  /* This had better be a $end.  
	     This should only happen once, the $end which matches $dumpvars   */
	  if (dollar_end_found) {
	      tf_message (ERR_WARNING, "VCDRead", "VCDRDL",
			  "'%s' found in VCD file, data changes section\n", prev_word);
	      tf_dofinish();
	  }
	  expect2("$end", prev_word);
	  dollar_end_found = 1;
      } /* $end */
      break; /* $ */


    case 'b': 
    case 'B': {
      /* Found a vector value - b[01xz]* abbrev */
      static char swap[MAX_LEN];  /* Used to swap bits */
      char *print_value;

      fscanf(vcdF, "%s %s ", value, abbrev);
      signal = abbrev_to_signal(abbrev);
      swap[0] = '\0';

      /* Schedule transistions on all signals with this abbreviation,
	 and which drive registers in the simulation.  */
      while (signal) {
	if (signal->argn) {
	    /* Is this a VSS signal that needs its bits swapped?           */
	    /* Cache the swapped value for multiple use                    */
	    if (vcdread_vss_bit_swap && !signal->bit_down) {  /* kg - add ! */
		if (!swap[0]) {
		    char pad;
		    int i, len = strlen(value);
		    for (i = 0; i < len; i++)
			swap[i] = value[(len-i)-1];

		    /* For short values, pad the upper bits with 0, Z, or X */
		    pad = value[0];
		    if (pad == '1') pad = '0';
		    for (i = len; i < signal->width; i++)
			swap[i] = pad;
		    swap[signal->width] = '\0';
		}
		print_value = swap;
	    }
	    else print_value = value;
	  if (verbose){
	    io_printf("VCD Driving %s (%s) = %s;\n", 
		   get_full_name(signal), abbrev, print_value);
	  }
	  tf_istrdelputp(signal->argn, signal->width, 'b', print_value,
			 signal->delay, INERTIAL_DELAY, signal->instance);
	}
	signal = signal->next_abbrev;
      } /* while (signal) */

      break; /* Binary vectors */
    } /* case 'b'*/

    case 'r':
    case 'R': {
      /* Found a real value - r999.999e99 abbrev */
      char *print_value;
      double rvalue;

      fscanf(vcdF, "%lf %s ", &rvalue, abbrev);
      signal = abbrev_to_signal(abbrev);

      /* Schedule transistions on all signals with this abbreviation,
	 and which drive registers in the simulation.  */
      while (signal) {
	if (signal->argn) {
	  if (verbose)
	    io_printf("VCD Driving %s (%s) = %e;\n", 
		   get_full_name(signal), abbrev, rvalue);
	  tf_iputrealp(signal->argn, rvalue, signal->instance);
	}
	signal = signal->next_abbrev;
      } /* while (signal) */

      break; /* Real vectors */
    } /* case 'r' */

    case '#':
      /* #delay are handled in vcdr_misc, which schedules delay */
      if (!dollar_end_found) {
	double time;
	get_vcd_time(&time);
	tf_message (ERR_ERROR, "VCDRead", "VCDRND",
		    "No $end found before first delay of #%0.0f'\n", 
		    time);
	tf_dofinish();
      }
      return TRUE;


    case 'p':
      /* This is only legal for MTI VCD files */
      if (use_std_abbrev) {
	tf_message (ERR_ERROR, "VCDRead", "VCDHuH",
		    "VCD Reader found '%c' in value changes - only legal with $vcdread_options(\"mti_abbrev\");", ch);
	tf_dofinish();
      }

      /* Funky MTI VHDL values #$*^@!
	 These look like 'pD 0 0 <999' where pD means '0', 0 0 means ???,
	 and <999 is the abbrev for signal 999 */
      fscanf(vcdF, "%c %d %d %s ", &level, &i, &j, abbrev);
      signal = abbrev_to_signal(abbrev);
      switch(level) {
      case 'D':      /* Down maps to 0 */
		value[0] = '0';	break;
      case 'L':      /* Low maps to 0 */
		value[0] = '0';	break;
      case 'U':      /* Up map to 1 */
		value[0] = '1';	break;
      case 'H':      /* High map to 1 */
		value[0] = '1';	break;
      case 'N':      /* N (undefined) maps to X */
		value[0] = 'X';	break;
      case '?':      /* ? (unknown) maps to X */
		value[0] = 'X';	break;
      case 'X':      /* X (don't care) maps to X */
		value[0] = 'X';	break;
      case 'Z':      /* Z (input three-state) maps to Z */
		value[0] = 'Z';	break;
      case 'T':      /* T (output three-state) maps to Z */
		value[0] = 'Z';	break;
      case 'f':      /* f means we are f***ed ??? bug ???*/
		value[0] = 'Z';	break;
      } /* switch(level) */
      value[1] = '\0';

      /* Schedule transistions on all signals with this abbreviation,
		 and which drive registers in the simulation.  */
      while (signal) {
	if (signal->argn) {
	  if (verbose){
	    io_printf("VCD Driving %s (%s) = %s (p%c);\n", 
		   get_full_name(signal), abbrev, value, level);
	  }
	  tf_istrdelputp(signal->argn, 1, 'b', value, 
			 signal->delay, INERTIAL_DELAY, signal->instance);
	}
	signal = signal->next_abbrev;
      }
      break;


    default:
      tf_message (ERR_ERROR, "VCDRead", "VCDHuH",
		  "VCD Reader found '%c' in value changes..ack!\n", ch);
      tf_dofinish();

    } /* switch (ch) */

    ch = fgetc(vcdF);
  } /* while (!feof(vcdF)) */

  if (!dollar_end_found) {
    tf_message (ERR_WARNING, "VCDRead", "VCDNED",
		"No $end found before end of file");
  }

  /* If there is a call to $vcdread_eof, set its argument to 1               */
  if (vcdr_eof_inst) {
    tf_iputp(ARG1, 1, vcdr_eof_inst);
  }

  /* The only way to get here is hitting the end of file - signal $vcdread */
  return FALSE;

} /* vcdread_changes */


/*********************************************************************/
/* Function to convert reason integer to reason name                 */
/*********************************************************************/
char *reason_name(int reason)
{
  char str[25];
  switch (reason) {
    case REASON_ENDOFCOMPILE : return("REASON_ENDOFCOMPILE"); break;
    case REASON_FINISH       : return("REASON_FINISH"); break;
    case REASON_INTERACTIVE  : return("REASON_INTERACTIVE"); break;
    case REASON_SYNCH        : return("REASON_SYNCH"); break;
    case REASON_ROSYNCH      : return("REASON_ROSYNCH"); break;
    case REASON_REACTIVATE   : return("REASON_REACTIVATE"); break;
    case REASON_PARAMVC      : return("REASON_PARAMVC"); break;
    case REASON_PARAMDRC     : return("REASON_PARAMDRC"); break;
    case REASON_SAVE         : return("REASON_SAVE"); break;
    case REASON_RESTART      : return("REASON_RESTART"); break;
    case REASON_RESET        : return("REASON_RESET"); break;
    case REASON_ENDOFRESET   : return("REASON_ENDOFRESET"); break;
    case REASON_FORCE        : return("REASON_FORCE"); break;
    case REASON_RELEASE      : return("REASON_RELEASE"); break;
 }
 return("Non-standard or Unknown Reason");
}
/*********************************************************************/


/******************************************************************************
  Procedure - 
      vcdr_misc

  Description - 
      Starts the whole stimulus thingie

      Value changes are read from the file and applied to simulation.
      Stop when a delay is read, and reschedule this routine for that time.

  Arguments - 
      data - not used
      reason - The usual PLI misc reasons
      At time 0, vcdr_call does a synchronize to wake up vcdr_misc
      with REASON_SYNCH.
      Rescheduling for a later time with tf_setdelay wakes up this
      routine with REASON_REACTIVATE

  Returns -
      void
******************************************************************************/
void vcdr_misc(int data, int reason)
{
  double next_vcd_time;

  switch (reason) {
  case REASON_SYNCH:
  case REASON_REACTIVATE:
    /* SYNCH is for the initial callback at time 0 while 
       REACTIVATE is when we get called because of tf_setdelay */

/*
    if (verbose) 
       io_printf("vcdr_misc: reason=%d (%s)\n", reason, reason_name(reason));
*/
    /* Read the next changes from the file.  If feof, don't look for time */
    if (vcdread_changes()) {
      get_vcd_time(&next_vcd_time);
      if (verbose) io_printf("VCD READ delay until @%0.0f\n", next_vcd_time);

      /* Wish list - this should all be done with 64 bit arithmetic */
      tf_setrealdelay(next_vcd_time * vcd2mod_scale - tf_getrealtime());
    }
      break;

  } /* switch (reason) */

} /* vcdr_misc */



/******************************************************************************
  Procedure - 
      vcdr_check_timescale

  Description - 
      Check the timescale read from the VCD file against this model's scale

  Arguments - 
      None

  Returns -
      void
******************************************************************************/
void vcdr_check_timescale()
{
  int multiplier, vcdUnit, 
    modUnit = tf_gettimeunit();
  char ch_unit[10];

  sscanf(vcd_timescale, "%d%s", &multiplier, ch_unit);

  if (!strcmp(ch_unit, "s")) vcdUnit = 0;
  else if (!strcmp(ch_unit, "ms")) vcdUnit = -3;
  else if (!strcmp(ch_unit, "us")) vcdUnit = -6;
  else if (!strcmp(ch_unit, "ns")) vcdUnit = -9;
  else if (!strcmp(ch_unit, "ps")) vcdUnit = -12;
  else if (!strcmp(ch_unit, "fs")) vcdUnit = -15;
  else {
    tf_message (ERR_WARNING, "VCDRead", "VCDSCA",
		"Unrecognized timescale '%s' in VCD file, using ns", 
		vcd_timescale);
    vcdUnit = -9;
  }

  if (multiplier == 10) vcdUnit = vcdUnit + 1;
  else if (multiplier == 100) vcdUnit = vcdUnit + 2;
  else if (multiplier != 1) {
    tf_message (ERR_WARNING, "VCDRead", "VCDSCA",
		"Unrecognized timescale multiplier '%s' in VCD file, using 1ns", 
		vcd_timescale);
    vcdUnit = -9;
  }

  /* Okay, we have the VCD Unit, compare it to the current module's unit */
  vcd2mod_scale = pow(10.0, (vcdUnit - modUnit));

  if (vcdUnit < modUnit) {
    /* VCD is smaller (ps) vs modules (ns), ie. scale < 1.0
       Pretty print the timescale and
       give a warning about possible rounding errors */
    char tscale[4], tunit[8], deci;
    if (modUnit < -12)     strcpy(tscale, "fs");
    else if (modUnit < -9) strcpy(tscale, "ps");
    else if (modUnit < -6) strcpy(tscale, "ns");
    else if (modUnit < -3) strcpy(tscale, "us");
    else if (modUnit < 0)  strcpy(tscale, "ms");
    else strcpy(tscale, "s");

    switch (modUnit % 3) {
    case -2: deci =  10; break;  // -2 =  10ms
    case -1: deci = 100; break;  // -1 = 100ms
    case  0: deci =   1; break;  //  0 =   1s
    case  1: deci =  10; break;  //  1 =  10s
    case  2: deci = 100; break;  //  2 = 100s
    default: tf_message (ERR_INTERNAL, "VCDRead", "VCDS",
		"Error computing timescale, mod value = %d", modUnit % 3);
    } /* switch */

    sprintf(tunit, "%d%s", deci, tscale);
    tf_message (ERR_WARNING, "VCDRead", "VCDSMS",
		"The VCD timescale (%s) is smaller than this module's timescale %s, rounding errors possible",
		  vcd_timescale, tunit);
  }

} /* vcdr_check_timescale */


/******************************************************************************
 $vcdread_signals routines

   Example

          $vcdread_signals(top.i1, top.o1);

   This says that value changes read in for top.i1 and top.o1 in the
   VCD file are to be applied to the signals top.i1 and top.o1 in the
   real simulation.  This is a quick way to specify stimulus points if
   they match between the VCD file and simulation.  The order of the
   signal list is not important.

   $vcdread_signals can only be called at time 0, after $vcdread has
   been called to open the VCD file and read the signal declarations.

******************************************************************************/

/******************************************************************************
  Procedure - 
      vcdr_sig_check

  Description - 
      Checks arguments for $vcdread_signals

  Arguments - 
      none

  Returns -
      void
******************************************************************************/
void vcdr_sig_check()
{
  int arg, nump = tf_nump();

  nump = tf_nump();
  if (nump == 0) {
    tf_message (ERR_ERROR, "Vcdr_Sig", "VRSP0",
		"$vcdread_signals must have at least one argument");
    return;
  }

  for (arg = 1; arg <= nump; arg++) {
    if (tf_typep(arg) != tf_readwrite) {
      tf_message (ERR_ERROR, "Vcdr_Sig", "VRSR",
		  "Argument %d to $vcdread_signals is not a register", arg);
      return;
    }
  }

} /* vcdr_sig_check */


/******************************************************************************
  Procedure - 
      vcdr_sig_call

  Description - 
      Build stimulus list of signal names for VCD and simulation 

  Arguments - 
      none

  Returns -
      void
******************************************************************************/
void vcdr_sig_call(int data, int reason)
{
  int arg;
  char *str;
  p_signal signal;

  /* Check $vcdread has been called yet and opened the VCD file */
  if (!vcdF) {
    tf_message (ERR_ERROR, "VCDRead", "VCDSCV",
	"$vcdread_signals called before $vcdread has opened the VCD file\n");
    tf_dofinish();
  }


  /* Currently all VCD setup must happen at time 0 */
  if (tf_gettime()) {
    tf_message (ERR_ERROR, "VCDRead", "VCDST0",
	"$vcdread_signals called after time 0\n");
    tf_dofinish();
  }

  /* Check the VCD timescale vs. this module's timescale */
  vcdr_check_timescale();

  /* Process the signal lists */
  arg = tf_nump();
  while (arg) {
    /* For each argument, get its name, and VCD signal block */
    str = acc_fetch_fullname(acc_handle_tfarg(arg));
    signal = str_to_signal(str);
    signal->argn = arg;
    signal->instance = tf_getinstance();
    signal->delay = signal_delay;
    arg--;
  } /* while arg */

} /* vcdr_sig_call */


/******************************************************************************
  $vcdread_signal_pairs routines

  Example:

   $vcdread_signal_pairs(top.in, "top.dut.in",  "top.dut.out", top.out);

   This says that simulation signal top.in will be driven by
   top.dut.in in the VCD file, and top.out will be driven by
   top.dut.out.  This call allows you to match up signals between the
   simulation and VCD file when they have different names.  The
   ordering of the signal pairs is not important, and within a pair,
   the VCD and simulation names can be in either order.

******************************************************************************/

/******************************************************************************
  Procedure - 
      vcdr_sigp_check

  Description - 
      Checks arguments for $vcdread_signal_pairs

  Arguments - 
      None

  Returns -
      void
******************************************************************************/
void vcdr_sigp_check()
{
  int arg,
    nump = tf_nump();

  if (nump < 2) {
    tf_message (ERR_ERROR, "Vcdr_Sig", "VRSP0",
		"$vcdread_signal_pairs must have at least two arguments");
    return;
  }

  if (nump % 2 != 0) {
    tf_message (ERR_ERROR, "Vcdr_Sigp", "VRSP2",
		"$vcdread_signal_pairs must an even number of arguments");
    return;
  }

  /* Look through pairs of arg for {reg,"str"} or {"str",reg}             */
  for (arg = 2; arg <= nump; arg=arg+2) {
    if (!
	(((tf_typep(arg)   == tf_readwrite) || (tf_typep(arg)   == tf_readwritereal)) &&
	 tf_typep(arg-1)   == tf_string) ||
	(((tf_typep(arg-1) == tf_readwrite) || (tf_typep(arg-1) == tf_readwritereal)) &&
	 tf_typep(arg)     == tf_string)) {
      tf_message (ERR_ERROR, "Vcdr_Sigp", "VRSP",
      "$vcdread_signal_pairs arguments %d & %d are not register, string pair",
		  arg-1, arg);
      if (tf_typep(arg) == tf_string)
	tf_message (ERR_ERROR, "VCDR", "YECH", "String was '%s'", tf_getcstringp(arg));
      else if (tf_typep(arg-1) == tf_string)
	tf_message (ERR_ERROR, "VCDR", "YECH", "String was '%s'", tf_getcstringp(arg-1));
      return;
    }
  } /* for arg */

} /* vcdr_sigp_check */


/******************************************************************************
  Procedure - 
      vcdr_sigp_call

  Description - 
      Build stimulus list of signal names for VCD and simulation 

  Arguments - 
      data - 1 = $vcdread_sigpnals, 2 = $vcdread_signal_pairs

  Returns -
      void
******************************************************************************/
void vcdr_sigp_call(int data, int reason)
{
  int arg;
  char *str;
  p_signal signal;

  /* Check $vcdread has been called yet and opened the VCD file */
  if (!vcdF) {
    tf_message (ERR_ERROR, "VCDRead", "VCDPCV",
	"$vcdread_signal_pairs called before $vcdread has opened the VCD file\n");
    tf_dofinish();
  }

  /* Currently all VCD setup must happen at time 0 */
  if (tf_gettime()) {
    tf_message (ERR_ERROR, "VCDRead", "VCDPT0",
	"$vcdread_signal_pairs called after time 0\n");
    tf_dofinish();
  }

  /* Check the VCD timescale vs. this module's timescale */
  vcdr_check_timescale();

  /* Process the signal lists */
  arg = tf_nump();
  while (arg) {

    /* For each pair of arguments, get sim name, and VCD name */
    if (tf_typep(arg) == tf_string) {
      /* arg-1 points to sim_name, arg points to "VCD_name" */
      str = tf_getcstringp(arg);
      signal = str_to_signal(str);
      signal->argn = arg-1;
    }
    else {
      /* arg-1 points to "VCD_name", arg points to sim_name */
      str = tf_getcstringp(arg-1);
      signal = str_to_signal(str);
      signal->argn = arg;
    }
    signal->instance = tf_getinstance();
    signal->delay = signal_delay;
    arg--; arg--;
  }

} /* vcdr_sigp_call */


/******************************************************************************
 $vcdread_clocks routines

   Example
	$vcdread_clocks(clock1, clock2, ...);

   This says that the listed signals are clocks, and that any other
   signal (specified with $vcdread_signals) should have a delay of 1.

   $vcdread_clocks can only be called at time 0, after $vcdread has
   been called to open the VCD file and read the signal declarations.

******************************************************************************/

/******************************************************************************
  Procedure - 
      vcdr_clocks_check

  Description - 
      Checks arguments for $vcdread_clocks

  Arguments - 
      none

  Returns -
      void
******************************************************************************/
void vcdr_clocks_check()
{
  int arg, nump = tf_nump();

  nump = tf_nump();
  if (nump == 0) {
    tf_message (ERR_ERROR, "Vcdr_Sig", "VRSP0",
		"$vcdread_clocks must have at least one argument");
    return;
  }

  for (arg = 1; arg <= nump; arg++) {
    if (tf_typep(arg) != tf_readwrite) {
      tf_message (ERR_ERROR, "Vcdr_Sig", "VRSR",
		  "Argument %d to $vcdread_clocks is not a register", arg);
      return;
    }
  }

  /* All signals (non-clock) will be applied with unit delay,
     unless signal_delay has already been set with $vcdoptions */
  if (signal_delay == ZERO_DELAY)
    signal_delay = UNIT_DELAY;

} /* vcdr_clocks_check */


/******************************************************************************
  Procedure - 
      vcdr_clocks_call

  Description - 
      Build stimulus list of clock names for VCD and simulation 

  Arguments - 
      none

  Returns -
      void
******************************************************************************/
void vcdr_clocks_call(int data, int reason)
{
  int arg;
  char *str;
  p_signal signal;

  /* Check $vcdread has been called yet and opened the VCD file */
  if (!vcdF) {
    tf_message (ERR_ERROR, "VCDRead", "VCDSCV",
	"$vcdread_clocks called before $vcdread has opened the VCD file\n");
    tf_dofinish();
  }

  /* Currently all VCD setup must happen at time 0 */
  if (tf_gettime()) {
    tf_message (ERR_ERROR, "VCDRead", "VCDST0",
	"$vcdread_clocks called after time 0\n");
    tf_dofinish();
  }

  /* Check the VCD timescale vs. this module's timescale */
  vcdr_check_timescale();

  /* Process the signal lists */
  arg = tf_nump();
  while (arg) {
    /* For each argument, get its name, and VCD signal block */
    str = acc_fetch_fullname(acc_handle_tfarg(arg));
    signal = str_to_signal(str);
    signal->argn = arg;
    signal->instance = tf_getinstance();
    signal->delay = ZERO_DELAY;
    arg--;
  } /* while arg */

} /* vcdr_clocks_call */


/******************************************************************************
  $vcdread_clock_pairs routines

  Example:

	$vcdread_clock_pairs(top.clock, "top.clkgen.clock");

   This says that simulation signal top.in will be driven by
   top.dut.in in the VCD file, and top.out will be driven by
   top.dut.out.  This call allows you to match up signals between the
   simulation and VCD file when they have different names.  The
   ordering of the signal pairs is not important, and within a pair,
   the VCD and simulation names can be in either order.

******************************************************************************/

/******************************************************************************
  Procedure - 
      vcdr_clockp_check

  Description - 
      Checks arguments for $vcdread_clock_pairs

  Arguments - 
      None

  Returns -
      void
******************************************************************************/
void vcdr_clockp_check()
{
  int arg,
    nump = tf_nump();

  if (nump < 2) {
    tf_message (ERR_ERROR, "Vcdr_Sig", "VRSP0",
		"$vcdread_clock_pairs must have at least two arguments");
    return;
  }

  if (nump % 2 != 0) {
    tf_message (ERR_ERROR, "Vcdr_Sigp", "VRSP2",
		"$vcdread_clock_pairs must an even number of arguments");
    return;
  }

  /* Look through pairs of arg for {reg,"str"} or {"str",reg}             */
  for (arg = 2; arg <= nump; arg=arg+2) {
    if ((tf_typep(arg)   != tf_readwrite ||
	 tf_typep(arg-1) != tf_string) &&
	(tf_typep(arg-1) != tf_readwrite ||
	 tf_typep(arg)   != tf_string)) {
      tf_message (ERR_ERROR, "Vcdr_Sigp", "VRSP",
      "$vcdread_clock_pairs arguments %d & %d are not register, string pair",
		  arg-1, arg);
      if (tf_typep(arg) == tf_string)
	tf_message (ERR_ERROR, "VCDR", "YECH", "String was '%s'", tf_getcstringp(arg));
      else if (tf_typep(arg-1) == tf_string)
	tf_message (ERR_ERROR, "VCDR", "YECH", "String was '%s'", tf_getcstringp(arg-1));
      return;
    }
  } /* for arg */

  /* All signals (non-clock) will be applied with unit delay */
  signal_delay = UNIT_DELAY;

} /* vcdr_clockp_check */


/******************************************************************************
  Procedure - 
      vcdr_clockp_call

  Description - 
      Build stimulus list of signal names for VCD and simulation 

  Arguments - 
      data - 1 = $vcdread_sigpnals, 2 = $vcdread_clock_pairs

  Returns -
      void
******************************************************************************/
void vcdr_clockp_call(int data, int reason)
{
  int arg;
  char *str;
  p_signal signal;

  /* Check $vcdread has been called yet and opened the VCD file */
  if (!vcdF) {
    tf_message (ERR_ERROR, "VCDRead", "VCDPCV",
	"$vcdread_clock_pairs called before $vcdread has opened the VCD file\n");
    tf_dofinish();
  }

  /* Currently all VCD setup must happen at time 0 */
  if (tf_gettime()) {
    tf_message (ERR_ERROR, "VCDRead", "VCDPT0",
	"$vcdread_clock_pairs called after time 0\n");
    tf_dofinish();
  }

  /* Check the VCD timescale vs. this module's timescale */
  vcdr_check_timescale();

  /* Process the signal lists */
  arg = tf_nump();
  while (arg) {

    /* For each pair of arguments, get sim name, and VCD name */
    if (tf_typep(arg) == tf_string) {
      /* arg-1 points to sim_name, arg points to "VCD_name" */
      str = tf_getcstringp(arg);
      signal = str_to_signal(str);
      signal->argn = arg-1;
    }
    else {
      /* arg-1 points to "VCD_name", arg points to sim_name */
      str = tf_getcstringp(arg-1);
      signal = str_to_signal(str);
      signal->argn = arg;
    }
    signal->delay = ZERO_DELAY;
    signal->instance = tf_getinstance();
    arg--; arg--;
  }

} /* vcdr_clockp_call */


/******************************************************************************
  $vcdread_options routines

  See vcdr_options_call for more info

  "verbose"         Turn on verbose mode
  "unverbose"       Turn off verbose mode [Default]
  "vss_bit_swap"    Swap bits for VSS files
  "novss_bit_swap"  Don't swap bits  [Default]
  "signal_delay=1"  Value to use in delaying signals (non-clocks)
  "mti_abbrev"	    Use MTI style abbreviations '<999'
  "bit_blast"       Turn vec[0], vec[1] into vec_0, vec_1
  "start_after_0"   Allow VCD files that start after time 0
******************************************************************************/

/******************************************************************************
  Procedure - 
      vcdr_options_check

  Description - 
      Checks arguments for $vcdread_options calls

  Arguments - 
      none used

  Returns -
      void
******************************************************************************/
void vcdr_options_check()
{
  int arg, nump = tf_nump();

  if (nump == 0) {
    tf_message (ERR_ERROR, "Vcdr_Options", "FILE1",
		"$vcdread_options must have at least one argument");
    return;
  }

  arg = nump;
  while (arg) {
    char *str;
    int delay;
    if (tf_typep(arg) != tf_string) {
      tf_message (ERR_ERROR, "Vcdr_Options", "ROBADA",
		  "Bad argument %d to $vcdread_options, must be string",
		  arg);
      return;
    }
    else if (!strncmp("signal_delay", str = tf_getcstringp(arg), 
		   strlen("signal_delay"))) {
	sscanf(str, "signal_delay=%d", &delay);
	signal_delay = delay;
    }

    arg --;
  } /* while arg */

} /* vcdr_options_check */


/******************************************************************************
  Procedure - 
      vcdr_options_call

  Description - 
      Gets the option strings and processes them.

  Arguments - 
      none used

  Returns -
      void
******************************************************************************/
void vcdr_options_call()
{
  int nargs, arg;
  char *str;

  /* Usually I process these arg lists backwards, but here I'm doing first
     to last so later args can reverse earlier - not bloodly likely */
  nargs = tf_nump();
  for (arg = 1; arg <= nargs; arg++) {
    str = tf_getcstringp(arg);

    if (!strcmp(str, "verbose")) 
      verbose = 1;
    else if (!strcmp(str, "unverbose")) 
      verbose = 0;
    else {
      if (vcdread_called) {
	tf_message (ERR_WARNING, "VCDRead", "VCDLate",
		    "Calls to $vcdread_options after $vcdread are too late to change how the file is read");
      }

      if (!strcmp(str, "vss_bit_swap"))
	vcdread_vss_bit_swap = TRUE;
      else if (!strcmp(str, "no_vss_bit_swap"))
	vcdread_vss_bit_swap = FALSE;
      else if (!strcmp(str, "mti_abbrev"))
	use_std_abbrev = FALSE;
      else if (!strcmp(str, "no_mti_abbrev"))
	use_std_abbrev = TRUE;
      else if (!strcmp(str, "bit_blast"))
	bit_blast = TRUE;
      else if (!strcmp(str, "no_bit_blast"))
	bit_blast = FALSE;
      else if (!strcmp(str, "start_after_0"))
	start_after_0 = TRUE;
      else if (!strncmp(str, "signal_delay", strlen("signal_delay"))) {
	if (verbose) {
	  int delay;
	  sscanf(str, "signal_delay=%d", &delay);
	  io_printf("Using signal_delay=%d from $vcdread_options\n", delay);
	}
      }
      else tf_message (ERR_WARNING, "Vcdr_Options", "Duh",
		       "Unrecognized argument %s to $vcdread_options",
		       str);
    }
  } /* for arg */

} /* vcdr_options_call */



/******************************************************************************
 $vcdread_eof routines
******************************************************************************/

/******************************************************************************
  Procedure - 
      vcdr_eof_check

  Description - 
      Checks arguments for $vcdread_eof calls

  Arguments - 
      none used

  Returns -
      void
******************************************************************************/
void vcdr_eof_check()
{
    if (tf_nump() != 1) {
      tf_message (ERR_ERROR, "VCDRead", "VCDFL1",
		  "$vcdread_eof has only one argument - the register to set on EOF");
      tf_dofinish();
    }

    /* Wish list - allow file name to be stored in a reg */
    if (tf_typep(ARG1) != tf_readwrite) {
      tf_message (ERR_ERROR, "VCDRead", "VCDSTR",
		  "Bad argument to $vcdread_eof, must be a reg");
      tf_dofinish();
    }

    /* Initialize this to 0 so $vcdread_eof can tell if $vcdread_eof
       has been called */
    vcdr_eof_inst = 0;

} /* vcdr_eof_check */


/******************************************************************************
  Procedure - 
      vcdr_eof_call

  Description - 
      Sets a pointer so that a user register can be set on EOF in the VCD file

  Arguments - 
      none used

  Returns -
      void
******************************************************************************/
void vcdr_eof_call()
{

   static bool vcdread_eof_called = {FALSE};

  /* Make sure $vcdread_eof is only called once during simulation */
  if (vcdread_eof_called) {
    tf_message (ERR_WARNING, "VCDRead", "VCDTWC", 
		"$vcdread_eof was called more than once during simulation,\n\tonly the last reg will be used");
  }
  else vcdread_eof_called = TRUE;

  /* Set up pointer to this instance so that on eof we can toggle the reg   */
  vcdr_eof_inst = tf_getinstance();

} /* vcdr_eof_call */
