/******************************************************************************
    vcdr_sa.c - VCD file reader
    These are the routines for the standalone version
    Copyright (C) 1998 Christian B. Spear and Synopsys, Inc.

    $Id: vcdr_sa.c,v 1.12 2008/04/17 20:08:22 spear Exp $
*/
int VCDR_SA_MAJOR_REV = 1;
int VCDR_SA_MINOR_REV = 10;
/*
Edit history
Who When      Version  What
cbs  7/17/02   1.10	Generate dummy Verilog code
cbs  1/03/01   1.9	Add from_vpd
cbs  8/08/00   1.8	Add support for real numbers
cbs  9/28/99   1.7	Add support for MTI style abbreviations
cbs  9/15/98   1.7	Add support for creating pattern files
cbs  9/14/98   1.6	Really pad the bits when reversing for VSS
cbs  8/03/98   1.5	Pad the bits when reversing for VSS
			Simulation time before $dumpvars is optional
cbs  7/31/98   1.3	Check fopen status
cbs  7/28/98   1.2	Add changes from Xerox for files created by VSS 
cbs  6/29/98   0.5	Lint cleanups
cbs  5/22/98   0.4	Added -s to search and print signal changes
cbs  5/16/98   0.3	Split into PLI & stand-alone versions
cbs  5/09/98   0.2	All pre-PLI stuff in place
cbs  3/22/98   0.1	Major data structures in place
cbs  3/17/98   0.0	Created

Wish list - Look for ???

- Should we write headers for pattern file?
- Should we write the final line of data?
- Change output format by modifying write_pattern_line

******************************************************************************/

extern int VCDR_MAJOR_REV, VCDR_MINOR_REV;


#include "malloc.h"        /* Reduces stress */
#include "string.h"        /* Better string definitions */

#define EXT extern         /* Make VCD data structures external */
#include "vcdr.h"          /* VCD Read data structures and globals */

extern bool vcdread_vss_bit_swap;
extern void exit(int);
extern void expect2(char *, char *);
extern char *get_full_name(p_signal);
extern void get_vcd_time(double *);

static double vtime = 0.0;  /* I'm cheating by making this a global.... */

/*****************************************************************************/
/* Signal find list.  This is a list of signals specified on the command */
/* line to be printed if found in the VCD file. */
/* Future uses would be do other actions on these signals */
typedef struct t_find
{
  char *name;              /* Name of signal to be found */
  struct t_find *next_find;/* Next signal to be found */
} s_find, *p_find;

p_find find_head = NULL;   /* Head of list of signals to be found */



/*****************************************************************************/
/* External routines found in vcdr.c */
extern vcdread_header();
extern vcdread_definitions();
extern expect(char *str);
extern p_signal abbrev_to_signal(char *abbrev);
extern p_signal str_to_signal (char *hname);


/******************************************************************************
  Procedure - 
      initialize

  Description - 
      Initializes global variables, prints version number

  Arguments - 
      argc, argv - Command line arguments

  Returns -
      void
******************************************************************************/
void initialize(int argc, char *argv[])
{
    int i;

    /* Set defaults */
    verbose = FALSE;
    strcpy(VCD_name, "verilog.dump");
    pattern = FALSE;
    pat_file[0] = '\0';
    name_file[0] = '\0';
    strobe_type = STROBE_NONE;
    strobe_name[0] = '\0';
    use_std_abbrev = TRUE; 	/* Use standard abbrevs, not MTI style */
    bit_blast = FALSE;          /* Don't bit blast vectors             */
    from_vpd = FALSE;           /* Most VCD files are not from vpd2vcd */
    gen_verilog = FALSE;

    /* Process the command line */
    /* Skip the first arg which is the program name, like vcdread */
    i = 1;
    while (i < argc) {
	if (verbose)
	    printf("Processing argv[%d]='%s'\n", i, argv[i]);
	
	if (!strcmp("-pat", argv[i])) {
	    i++;
	    if (i >= argc) {
		printf("ERROR, No pattern file name given with -pat\n");
		exit(1);
	    }
	    strcpy (pat_file, argv[i]);
	    pattern = TRUE;
	    if (verbose) 
		printf("-pat %s found!\n", pat_file);
	} /* -pat */

	else if (!strcmp("-pname", argv[i])) {
	    i++;
	    if (i >= argc) {
		printf("ERROR, No pattern name file given with -pname\n");
		exit(1);
	    }
	    strcpy (name_file, argv[i]);
	    pattern = TRUE;
	    if (verbose) printf("-pname %s found!\n", name_file);
	} /* -pname */

	else if (!strcmp("-ptime", argv[i])) {
	    if (strobe_type != STROBE_NONE) {
	        printf("ERROR, only one of -ptime and -pstrobe can be given!\n");
		exit(1);
	    }
	    pattern = TRUE;
	    strobe_type = STROBE_TIME;
	    if (verbose) printf("-ptime found!\n");
	} /* -ptime */

	else if (!strcmp("-pstrobe", argv[i])) {
	    if (strobe_type != STROBE_NONE) {
	        printf("ERROR, only one of -ptime and -pstrobe can be given!\n");
		exit(1);
	    }
	    pattern = TRUE;
	    strobe_type = STROBE_SIGNAL;
	    i++;
	    if (i >= argc) {
		printf("ERROR, No signal name given with -pstrobe\n");
		exit(1);
	    }
	    strcpy (strobe_name, argv[i]);
	    if (verbose) printf("-pstrobe %s found!\n", strobe_name);
	    printf("ERROR, -pstrobe not currently implemented\n");
	    exit(1);
	} /* -pstrobe */

	else if (!strcmp("-pradix=2", argv[i])) {
	    pattern = TRUE;
	    if (verbose) printf("-pradix=2\n");
	} /* -pradix=2 */

	else if (!strcmp("-pradix=16", argv[i])) {
	    pattern = TRUE;
	    printf("WARNING, -pradix=16 unsupported and ignored\n");
	} /* pradix=16 */

	else if (!strcmp("-verilog", argv[i])) {
	    gen_verilog = TRUE;
	    if (verbose) printf("-verilog found!\n");
	} /* -verilog */

	else if (!strcmp("-v", argv[i]))
	    verbose = TRUE;

	else if (!strcmp("-V", argv[i]))
	    verbose = FALSE;

	else if (!strcmp("-s", argv[i])) {
	    /* Find this signal in the VCD file and print its value         */
	    /* Create a new element in the linked list of signal names and  */
	    /* save the signal name                                         */
	    p_find find;
	    find = malloc(sizeof(s_find));
	    memset((void *)find, sizeof(s_find), 0);
	    i++;
	    find->name = strcpy ((char *)malloc(strlen(argv[i]) + 1), argv[i]);
	    find->next_find = find_head;
	    find_head = find;
	    if (verbose)
		printf("\tFind signal '%s'\n", find->name);

	} /* if -s */

	else if (!strcmp("-vss_bit_swap", argv[i]))
	    vcdread_vss_bit_swap = TRUE;
	else if (!strcmp("-no_vss_bit_swap", argv[i]))
	    vcdread_vss_bit_swap = FALSE;

	else if (!strcmp("-mti_abbrev", argv[i]))
		use_std_abbrev = FALSE;
	else if (!strcmp("-no_mti_abbrev", argv[i]))
		use_std_abbrev = TRUE;

	else if (!strcmp("-bit_blast", argv[i]))
		bit_blast = TRUE;
	else if (!strcmp("-no_bit_blast", argv[i]))
		bit_blast = FALSE;

	else if (!strcmp("-from_vpd", argv[i]))
		from_vpd = TRUE;
	else if (!strcmp("-not_from_vpd", argv[i]))
		from_vpd = FALSE;

	/* Give help if -h, -H, -? or any unknown switch                     */
	else if (! strcmp("-h", argv[i]) | 
		 ! strcmp("-H", argv[i]) |
		 ! strcmp("-?", argv[i]) |
		 ((argv[i])[0] == '-')) {
	    printf ("Usage:\t%s [options] [VCD_file]\n", argv[0]);
	    printf ("\tOptions:\n");
	    printf ("\t\t[-pat output.pat  (Create a pattern file)]\n");
	    printf ("\t\t[-pname names.txt (Names of signals in patterns)]\n");
	    printf ("\t\t[-pradix=[2,16]    (Radix for pattern file, only 2 supported)]\n");
	    printf ("\t\t[-ptime            (Write pattern when time change)]\n");
	    printf ("\t\t[-pstrobe signal   (Write a pattern when strobe signal changes - not supported)]\n");
	    printf ("\t\t[-s signal (print signal values)]\n");
	    printf ("\t\t[-verilog (generate Verilog shell)]\n");
	    printf ("\t\t[-v (verbose)]\n");
	    printf ("\t\t[-V (quiet)]\n");
	    printf ("\t\t[-vss_bit_swap (swap bits on $vars with [hi:lo])\n");
	    printf ("\t\t[-no_vss_bit_swap (don't swap - default)\n");
	    printf ("\t\t[-from_vpd (file produced by vpd2vcd)\n");
	    printf ("\tDefault VCD file name is verilog.dump\n");
	    exit(0);
	} /* if help */

	/* Default is the VCD file name                                      */
	else strcpy(VCD_name, argv[i]);

	i++;
    } /* while i */

    if (verbose) 
	printf("VCD Read v%d.%d(vcdr_sa.c) + v%d.%d(vcdr.h) + v%d.%d(vcdr.c)\n", 
	       VCDR_SA_MAJOR_REV, VCDR_SA_MINOR_REV,
	       VCDR_H_MAJOR_REV,  VCDR_H_MINOR_REV,
	       VCDR_MAJOR_REV, VCDR_MINOR_REV);


    /* Set up for pattern files - actions that cen be done before VCD is
       read */
    if (pattern) {
	if (strobe_type == STROBE_NONE) {
	    strobe_type = STROBE_TIME;
	}

	if (pat_file[0] == '\0') {
	    printf("ERROR, No pattern file name given with -pat switch\n");
	    exit(1);
	}

	if (name_file[0] == '\0') {
	    printf("ERROR, No signal name file name given with -pname switch\n");
	    exit(1);
	}

    } /* if pattern */

} /* initialize */


/******************************************************************************
  Procedure - 
  find_signals
  
  Description - 
  Find the signals specified on the command line with -s and mark them
  to be printed when their changes are found in the VCD file
  
  Arguments - 
  None
  
  Returns -
  nothing
  ******************************************************************************/
void find_signals()
{
    p_find find;
    
    find = find_head;
    while (find) {
	p_signal signal;
	signal = str_to_signal(find->name);
	if (signal)
	    signal->print = 1;
	else
	    printf("WARNING, The signal '%s' was not found in the VCD file.\n",
		   find->name);
	find = find->next_find;
    }

/******************************************************************************
    printf("Found signal '%s' when looking up '%s'\n",
	   get_full_name (str_to_signal ("tb_rtl.b")), "tb_rtl.b");

    printf("Found signal '%s' when looking up '%s'\n",
	   get_full_name (str_to_signal ("2top.wtop")), 
	   "2top.wtop");

    printf("Found signal '%s' when looking up '%s'\n",
	   get_full_name (str_to_signal ("2top.m1.b3.3bot")), 
	   "2top.m1.b3.3bot");
******************************************************************************/

    
} /* find_signals */


/******************************************************************************
  Procedure - 
      pattern_open

  Description - 
      Opens and reads the name file.  If this is successful,
      opens the pattern file.

  Arguments - 
      None

  Returns -
      nothing
******************************************************************************/
void pattern_open()
{
    char line[MAX_LEN], name[MAX_LEN];
    p_signal signal;

    if (verbose) printf("Opening pattern signal name file %s\n", name_file);
    nameF = fopen(name_file,"r");
    if (!nameF) {
	printf("ERROR, unable to open the pattern signal name file '%s'\n",
	       name_file);
	exit (1);
    }

    head_pnames = NULL;
    tail_pnames = NULL;

    /* Read the signal names from the file, marking the signals as special! */
    fgets(line, MAX_LEN, nameF);
    while (!feof(nameF)) {
	
	sscanf(line, "%s", name);
	
	/* Strip off comments */
	if (strncmp("//", name, 2) && 
	    strncmp("--", name, 2)) {
	    p_signal signal = str_to_signal(name);
	    if (signal) {
		if (verbose) printf("Read %s from pattern name file\n", name);
		if (signal->width == 1) {
		    /* This is a scalar signal, hold prev value in char */
		    signal->last_val.scalar = 'x';
		}
		else {
		    /* This is vector val, malloc a place to hold last val */
		    int i;
		    char *vector;
	    
		    vector = (char *) malloc(signal->width + 1);
		    signal->last_val.vector = vector;
		    for (i=signal->width-1; i<0; i--)
			vector[i] = 'x';
		    vector[signal->width] = '\0';		 
		} /* if signal width == 1 */

		if (head_pnames) {
		    /* Link new signal to the end of the list */
		    tail_pnames->next_psignal = signal;
		    tail_pnames = signal;
		}
		else {
		    /* Empty list, add first signal */
		    head_pnames = signal;
		    tail_pnames = signal;
		} /* if head_pnames */
	    }
	    else {
		printf("ERROR, signal name '%s' from pattern name file was not found\n",
		       name);
		exit(1);
	    } /* if signal */
	    
	} /* if comment */

	fgets(line, MAX_LEN, nameF);
    } /* while !eof nameF */

    fclose(nameF);


    if (verbose) printf("Opening pattern file %s\n", pat_file);
    patF = fopen(pat_file,"w");
    if (!patF) {
	printf("ERROR, unable to open the pattern file '%s'\n", pat_file);
	exit (1);
    }

    /* ??? Should we write a header?  Maybe even the signals from above? */
    

} /* pattern_open */


/******************************************************************************
  Procedure - 
      write_pattern_line
  
  Description - 
      Writes a single line to the pattern file for the given time
  
  Arguments - 
      none
  
  Returns -
      nothing
******************************************************************************/
void write_pattern_line()
{
    p_signal signal;
    
    fprintf(patF, "%0.0f ", vtime);
    
    signal = head_pnames;
    while (signal) {
	if (signal->width == 1) 
	    fprintf(patF, "%c ", signal->last_val.scalar);
	else
	    fprintf(patF, "%s ", signal->last_val.vector);
	signal = signal->next_psignal;
    } /* while signal */

    fprintf(patF, "\n");

} /* write_pattern_line */


/******************************************************************************
  Procedure - 
      pattern_close
  
  Description - 
      Writes the final pattern and closes pattern file
  
  Arguments - 
      None
  
  Returns -
      nothing
  ******************************************************************************/
void pattern_close()
{
    /* Write the final pattern ???? */
    write_pattern_line();
    
    /* Close the bloody pattern file already! */
    fclose(patF);

    if (verbose) printf("Successfully created pattern file %s", pat_file);
} /* pattern_close */


/******************************************************************************
  Procedure - 
      vcdread_changes

  Description - 
      Reads the signal changes from the VCD file

  Arguments - 
      None

  Returns -
      Nothing
******************************************************************************/
void vcdread_changes()
{
    char abbrev[MAX_ABBREV_LEN], value[MAX_LEN], ch, level;
    p_signal signal;
    bool dollar_end_found;
    int i, j;

    /* Read initial signal values, then later values
       #0         Optional in XL
       $dumpvars
       <Initial values>
       $end
       #<delay>
       <More values>
       <EOF>
    */
    if (verbose) printf("Initial values: ");

    /* #0 is optional in XL dump files */
    fscanf(vcdF, "%s ", prev_word);
    if (! strcmp(prev_word, "#0")) {
	fscanf(vcdF, "%s ", prev_word);
    }
    expect2("$dumpvars", prev_word);


    dollar_end_found = 0;
    ch = fgetc(vcdF);
    while(!feof(vcdF)) {
        switch (ch) {

	case '0': 
	case '1': 
	case 'z': 
	case 'x': 
	case 'Z': 
	case 'X': 
	    /* Found a scalar value - [01xz]abbrev */
	  fscanf(vcdF, "%s ", abbrev);

	  if (!strcmp(abbrev, "s"))
	      ch = 's';
	  signal = abbrev_to_signal(abbrev);

	  /* Print out all the signals with this abbreviation                */
	  /* Print the time if the signal was given with -s on the cmd line  */
	  while (signal) {
	      if (verbose){
		printf("%s = %c; // %s\n", get_full_name(signal), ch, abbrev);
	      }
	      else if (signal->print){
		printf("@%0.0f. %s = %c; // %s\n", 
		       vtime, get_full_name(signal), ch, abbrev);
	      }
	      signal->last_val.scalar = ch; /* Save the last value */
	      signal = signal->next_abbrev;
	  }
	  break; /* 01xzXZ */


	case '$':
	  /* This should only happen once, the $end which matches $dumpvars  */
	  ungetc(ch, vcdF);
	  if (dollar_end_found) {
	    char line[MAX_LEN];
	    fscanf(vcdF, "%s ", line);
	    printf("ERROR, '%s' found in VCD file changes section\n", line);
	    exit(1);
	  }
	  expect("$end");
	  dollar_end_found = 1;
	  break;


	case 'b':
	case 'B': {
	  /* Found a vector value - b[01xz] abbrev */
	  static char swap[MAX_LEN] = {""};  /* Used to swap bits */
	  char *print_value;

	  fscanf(vcdF, "%s %s ", value, abbrev);
	  signal = abbrev_to_signal(abbrev);

	  /* Print out all the signals with this abbreviation                */
	  while (signal) {

	      /* Is this a VSS signal that needs its bits swapped?           */
	      /* Cache the swapped value for multiple use                    */
	      if (vcdread_vss_bit_swap & (signal->bit_down)) {
		  if (!swap[0]) {
		      char pad;
		      int i, len = strlen(value);
		      for (i = 0; i < len; i++)
			  swap[i] = value[(len-i)-1];

		      /* For short values, pad with 0, Z, or X               */
		      pad = value[0];
		      if (pad == '1') pad = '0';
		      for (i = len; i < signal->width; i++)
			  swap[i] = pad;
		      swap[signal->width] = '\0';
		  }
		  print_value = swap;
	      }
	      else print_value = value;

	      if (verbose)
		  printf("%s = %s; // %s\n",
			 get_full_name(signal), print_value, abbrev);
	      else if (signal->print)
		  printf("@%0.0f. %s = %s; // %s\n",
			 vtime, get_full_name(signal), print_value, abbrev);

	      /* Save the last value for pattern files */
	      if (pattern && signal->last_val.vector) 
		  strcpy(signal->last_val.vector, print_value);

	      signal = signal->next_abbrev;
	  } /* while signal */
	  swap[0] = '\0';
	  break;
	} /* case 'b' */


	case 'r':
	case 'R': {
	  /* Found a real value - r999.999e99 abbrev */
	  char *print_value;
	  double rvalue;

	  fscanf(vcdF, "%lf %s ", &rvalue, abbrev);
	  signal = abbrev_to_signal(abbrev);

	  /* Schedule transistions on all signals with this abbreviation,
	     and which drive registers in the simulation.  */
	  while (signal) {
	    if (verbose)
	      printf("%s = %e; // %s\n", 
		     get_full_name(signal), rvalue, abbrev);
	    else if (signal->print)
	      printf("@%0.0f. %s = %e; // %s\n",
		     vtime, get_full_name(signal), rvalue, abbrev);

	      /* Save the last value for pattern files */
	      if (pattern && signal->last_val.vector) 
		  signal->last_val.real = rvalue;

	    signal = signal->next_abbrev;
	  } /* while (signal) */

	  break; /* Real vectors */
	} /* case 'r' */



	case '#':
	  /* This is a new time - write a line to the patter file? */
	  if (pattern && strobe_type == STROBE_TIME) 
	      write_pattern_line();

	  get_vcd_time(&vtime);
	  if (!dollar_end_found) {
	      printf("ERROR, No $end found before first delay of %0.0f'\n", 
		     vtime);
	      exit(1);
	  }
	  if (verbose) printf("@%0.0f: ", vtime);
	  break;


    case 'p':
      /* This is only legal for MTI VCD files */
      if (use_std_abbrev) {
	printf ("ERROR, VCD Reader found '%c' in value changes - only legal with $vcdread_options(\"mti_abbrev\");\n", ch);
	exit(1);
      }

      /* Funky MTI VHDL values #$*^@!
	 These look like 'pD 0 0 <999' where pD means '0', 0 0 means ???,
	 and <999 is the abbrev for signal 999 */
      fscanf(vcdF, "%c %d %d %s ", &level, &i, &j, abbrev);
      signal = abbrev_to_signal(abbrev);
      switch(level) {
      case 'D':      /* Down maps to 0 */
		value[0] = '0';	break;
      case 'L':      /* Low maps to 0 */
		value[0] = '0';	break;
      case 'U':      /* Up map to 1 */
		value[0] = '1';	break;
      case 'H':      /* High map to 1 */
		value[0] = '1';	break;
      case 'N':      /* N (undefined) maps to X */
		value[0] = 'X';	break;
      case '?':      /* ? (unknown) maps to X */
		value[0] = 'X';	break;
      case 'X':      /* X (don't care) maps to X */
		value[0] = 'X';	break;
      case 'Z':      /* Z (input three-state) maps to Z */
		value[0] = 'Z';	break;
      case 'T':      /* T (output three-state) maps to Z */
		value[0] = 'Z';	break;
      case 'f':      /* f means we are f***ed ??? bug ???*/
		value[0] = 'Z';	break;
      } /* switch(level) */
      value[1] = '\0';

      /* Schedule transistions on all signals with this abbreviation,
	 and which drive registers in the simulation.  */
      while (signal) {
	  if (verbose){
	    printf("%s = %c (p%c); // %s\n", 
		   get_full_name(signal), value[0], level, abbrev);
	  }
	  else if (signal->print){
	    printf("@%0.0f. %s = %c (p%c); // %s\n", 
		   vtime, get_full_name(signal), value[0], level, abbrev);

	  }
	  signal = signal->next_abbrev;
      } // while (signal) 
      break;





	default:
	  printf("ERROR, VCD Reader found '%c' in value changes..ack!\n", ch);
	  exit(1);


	} /* switch (ch) */


	ch = fgetc(vcdF);
    } /* while (!feof(vcdF)) */

    if (!dollar_end_found) {
      printf("ERROR, No $end found before end of file");
      exit(1);
    }

} /* vcdread_changes */


/******************************************************************************
  Procedure - 
      vcdread_file

  Description - 
      Opens and reads the VCD file

  Arguments - 
      None

  Returns -
      nothing
******************************************************************************/
void vcdread_file()
{
    if (verbose) printf("Opening VCD file %s\n", VCD_name);
    vcdF = fopen(VCD_name,"r");
    if (!vcdF) {
	printf("ERROR, unable to open the VCD file '%s'\n", VCD_name);
	exit (1);
    }
    vcdread_header();
    vcdread_definitions();
    if (pattern) pattern_open();
    find_signals();
    vcdread_changes();
    if (pattern) pattern_close();
    fclose(vcdF);

} /* vcdread_file */


/******************************************************************************
  Procedure - 
      main

  Description - 
      King of the hill, top of the heap

  Arguments - 
      None

  Returns -
      True
******************************************************************************/
int main(int argc, char *argv[])
{
  initialize(argc, argv);

  vcdread_file();

  if (verbose) printf("That's all folks!\n");

  return 0;
}
